# util.py - various shared utility functions
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os, struct
from shutil import rmtree
from subprocess import Popen, CalledProcessError, PIPE, STDOUT
from pipes import quote as shellquote
from redhat_upgrade_tool import grub_conf_file

import logging
log = logging.getLogger(__package__+".util")

try:
    from ctypes import cdll, c_bool
    selinux = cdll.LoadLibrary("libselinux.so.1")
    is_selinux_enabled = selinux.is_selinux_enabled
    is_selinux_enabled.restype = c_bool
except (ImportError, AttributeError, OSError):
    is_selinux_enabled = lambda: False

def call(*popenargs, **kwargs):
    return Popen(*popenargs, **kwargs).wait()

def check_output(*popenargs, **kwargs):
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd)
    return output

def check_call(*popenargs, **kwargs):
    retcode = call(*popenargs, **kwargs)
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd)
    return 0

def listdir(d):
    for f in os.listdir(d):
        yield os.path.join(d, f)

def rlistdir(d):
    for root, dirs, files in os.walk(d):
        for f in files:
            yield os.path.join(root, f)

def mkdir_p(d):
    try:
        os.makedirs(d)
    except OSError as e:
        if e.errno != 17:
            raise

def rm_f(f, rm=os.remove):
    if not os.path.lexists(f):
        return
    try:
        rm(f)
    except (IOError, OSError) as e:
        log.warn("failed to remove %s: %s", f, str(e))

def rm_rf(d):
    if os.path.isdir(d):
        rm_f(d, rm=rmtree)
    else:
        rm_f(d)

def kernelver(filename):
    '''read the version number out of a vmlinuz file.'''
    # this algorithm came from /usr/share/magic
    with open(filename) as f:
        f.seek(514)
        if f.read(4) != 'HdrS':
            return None
        f.seek(526)
        (offset,) = struct.unpack("<H", f.read(2))
        f.seek(offset+0x200)
        buf = f.read(256)
    uname, nul, rest = buf.partition('\0')
    version, spc, rest = uname.partition(' ')
    return version

def df(mnt, reserved=False):
    s = os.statvfs(mnt)
    return s.f_bsize * (s.f_bfree if reserved else s.f_bavail)

def hrsize(size, si=False, use_ib=False):
    powers = 'KMGTPEZY'
    multiple = 1000 if si else 1024
    if si:       suffix = 'B'
    elif use_ib: suffix = 'iB'
    else:        suffix = ''
    size = float(size)
    for p in powers:
        size /= multiple
        if size < multiple:
            if p in 'KM': # don't bother with sub-MB precision
                return "%u%s%s" % (int(size)+1, p, suffix)
            else:
                return "%.1f%s%s" % (size, p, suffix)



# callback.py - callback functions for progress reporting
# vim: set fileencoding=UTF-8:
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import rpm
import logging
from rpmUtils.miscutils import formatRequire
from yum.callbacks import ProcessTransBaseCallback

def format_pkgtup(tup):
    (n,a,e,v,r) = tup
    if e not in (0, '0', None):
        return '%s:%s-%s-%s.%s' % (e,n,v,r,a)
    else:
        return '%s-%s-%s.%s' % (n,v,r,a)


# callback objects for RPM transactions

class BaseTsCallback(object):
    '''Basic RPMTransaction Callback class. You need one of these to actually
       make a transaction work.
       If you subclass it, make sure you preserve the behavior of
       inst_open_file and inst_close_file, or nothing will actually happen.'''
    callback_map = dict((rpm.__dict__[k], k[12:].lower())
                         for k in rpm.__dict__
                         if k.startswith('RPMCALLBACK_'))

    def __init__(self):
        self._openfds = dict()
        self.log = logging.getLogger(__package__+".rpm")

    def callback(self, what, amount, total, key, data):
        if what not in self.callback_map:
            self.log.info("Ignoring unknown callback number %i", what)
            return
        name = self.callback_map[what]
        #self.log.debug("%s(%s, %s, %s, %s)", name, amount, total, key, data)
        func = getattr(self, name, None)
        if callable(func):
            return func(amount, total, key, data)

    def openfile(self, filename):
        f = open(filename, 'r')
        self._openfds[filename] = f
        return f.fileno()

    def closefile(self, filename, unlink=False):
        f = self._openfds.pop(filename)
        f.close()
        if unlink:
            os.unlink(filename)

    def inst_open_file(self, amount, total, key, data):
        '''Called whenever RPM wants to open a package file.
           The key will be a filename, or txmbr object, or whatever you passed
           as the 'key' argument to ts.addInstall().
           Must return a fd for the now-open file.'''
        raise NotImplementedError

    def inst_close_file(self, amount, total, key, data):
        '''Called when RPM wants to close a previously-opened file.
        'key' will be the same as the 'inst_open_file' call.'''
        raise NotImplementedError

class RPMTsCallback(BaseTsCallback):
    '''Minimal RPM transaction callback that opens/closes files as required'''
    def inst_open_file(self, amount, total, filename, data):
        return self.openfile(filename)

    def inst_close_file(self, amount, total, filename, data):
        self.closefile(filename)

class DownloadCallbackBase(ProcessTransBaseCallback):
    def __init__(self):
        ProcessTransBaseCallback.__init__(self)
        self.log = logging.getLogger(__package__+".download")

    # for Yum transaction callbacks (i.e. YumBase.processTransaction stuff)
    def event(self, state, data=None):
        ProcessTransBaseCallback.event(self, state, data)

    # our 'verify' callback in download_packages (yum doesn't have one :/)
    def verify(self, amount, total, filename, data):
        shortname = filename.split('/')[-1]
        self.log.debug("verifying %u/%u %s", amount, total, shortname)

# callback object for depsolving

class DepsolveCallbackBase(object):
    modedict = {
        'i': 'install',
        'u': 'update',
        'e': 'erase',
        'r': 'reinstall',
        'd': 'downgrade',
        'o': 'obsolete',
        'ud': 'updated',
        'od': 'obsoleted',
    }
    def __init__(self, yumobj=None):
        self.yumobj = yumobj
        if yumobj:
            self.yum_setup(yumobj)
        self.log = logging.getLogger(__package__+".depsolve")
        self.mode_counter = dict((m, 0) for m in self.modedict)
        self.missingreqs = set()
    def yum_setup(self, yumobj):
        pkglist = yumobj.doPackageLists(pkgnarrow='installed')
        self.installed_packages = len(pkglist.installed)
    def start(self):
        self.log.debug('starting depsolve')
    def tscheck(self):
        self.log.debug('running transaction check')
    def restartLoop(self):
        self.log.debug('restarting depsolve')
    def end(self):
        self.log.debug('finished depsolve')
        self.log.debug('%u updates for %u packages',
                      self.mode_counter['u'], self.installed_packages)
        self.log.debug('%s', self.mode_counter)
        self.log.debug("%u missing reqs", len(self.missingreqs))
        for tup in self.missingreqs:
            self.log.debug("missing: %s", formatRequire(*tup))
    def pkgAdded(self, tup, mode):
        self.mode_counter[mode] += 1
        pkg = format_pkgtup(tup)
        if mode in ('e', 'od', 'ud'):
            self.log.debug('%s: remove (%s)',pkg, self.modedict[mode])
        elif mode in ('u', 'i', 'd', 'r', 'o'):
            self.log.debug('%s: install (%s)', pkg, self.modedict[mode])
    def procReqPo(self, po, formatted_req):
        self.log.debug('%s → %s', po, formatted_req)
    def procConflictPo(self, po, formatted_conflict):
        self.log.debug('CONFLICT: %s → %s', po, formatted_conflict)
    def unresolved(self, msg):
        self.log.debug('UNRESOLVED DEP: %s', msg)
    def format_missing_requires(self, po, tup):
        req = formatRequire(*tup)
        self.log.debug('MISSING REQ: %s requires %s', po, req)



# repofile.py - alters .repo files (adds or replaces repo options)
#
# Copyright (C) 2016 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import re


class RepoFileParser(object):
    def __init__(self, repo_file):
        self._repo_file = repo_file
        self._repo_file_content = self._get_repo_file_content()
        self._repo_ids = self._get_repo_ids()

    def _get_repo_file_content(self):
        with open(self._repo_file) as file_to_read:
            return file_to_read.read()

    def _get_repo_ids(self):
        """Retrieve IDs of all the repositories within the .repo file."""
        return re.findall("^\s*\[(.*?)\]\s*$",
                          self._repo_file_content,
                          re.MULTILINE)

    def set_option(self, option_name, value_to_set, orig_value=None):
        """Go through all the the repo IDs (sections) within the repo file
        and set new value of a specific repository option. The option is
        added if it does not exist. Optional parameter orig_value causes that
        the repository option value is replaced only if the current value
        matches the orig_value.
        """
        for repo_id in self._repo_ids:
            # Search for the option among repo ID options
            current_value = self._get_option_value(repo_id, option_name)
            if current_value:
                if orig_value and current_value != orig_value:
                    continue
                self._replace_option_value(repo_id, option_name, value_to_set)
            elif not current_value:
                # The option isn't present - add it
                self._add_option(repo_id, option_name, value_to_set)

    def _get_option_value(self, repo_id, option_name):
        """Return value of an option present under a specific repo ID.
        Return None if the option is not found.
        """
        option_found = re.search(
            r"\[{0}\][^\[]*?{1}\s*=\s*(\S+)\s*$"
            .format(repo_id, option_name),
            self._repo_file_content, flags=re.MULTILINE | re.DOTALL)
        return option_found.group(1) if option_found else None

    def _replace_option_value(self, repo_id, option_name, option_value):
        pattern = re.compile(
            r"(\[{0}\][^\[]*?{1}\s*=\s*).*?$".format(repo_id, option_name),
            flags=re.MULTILINE)
        self._repo_file_content = pattern.sub(
            r"\g<1>{0}".format(option_value), self._repo_file_content)

    def _add_option(self, repo_id, option_name, option_value):
        """Add a new option after the baseurl option within the section of
        a specific repo ID.
        """
        pattern = re.compile(r"(\[{0}\].*?baseurl.*?)$".format(repo_id),
                             flags=re.MULTILINE | re.DOTALL)
        self._repo_file_content = pattern.sub(
            r"\1\n{0}={1}".format(option_name, option_value),
            self._repo_file_content)

    def write(self):
        """Write the altered repo file content to the original file."""
        with open(self._repo_file, 'w') as file_to_write:
            file_to_write.write(self._repo_file_content)



# media.py - check for installable media, mount/umount devices, etc.
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os, stat, re
from collections import namedtuple
from os.path import exists, join, realpath
from .util import check_output, call, STDOUT, CalledProcessError
from tempfile import mkdtemp

import logging
log = logging.getLogger(__package__+".media")

def check_call(cmd, stderr=STDOUT, *args, **kwargs):
    # Use stderr=STDOUT so CalledProcessError has mount output in e.output
    check_output(cmd, stderr=stderr, *args, **kwargs)

class FstabEntry(namedtuple('FstabEntry','dev rawmnt type opts freq passno')):
    __slots__ = ()
    @property
    def mnt(self):
        '''unescaped mountpoint'''
        return self.rawmnt.decode('string_escape')

def mounts(fstab="/proc/mounts"):
    for line in open(fstab):
        if line.strip().startswith('#') or line.isspace():
            continue
        yield FstabEntry(*line.split())

def ismedia(mountpoint):
    return exists(join(mountpoint, ".treeinfo"))

def isblock(dev):
    return exists(dev) and stat.S_ISBLK(os.stat(dev).st_mode)

def iscd(dev):
    typefile = "/sys/class/block/%s/device/type" % os.path.basename(dev)
    return os.path.exists(typefile) and int(open(typefile).read()) == 5

def isiso(filename):
    try:
        iso = open(filename)
        iso.seek(32769)
        magic = iso.read(5)
    except IOError:
        magic = ''
    return magic == 'CD001'

def isloop(dev):
    return exists(dev) and os.major(os.stat(dev).st_rdev) == 7

def find():
    return [m for m in mounts() if isblock(m.dev) and ismedia(m.mnt)]

def removable():
    '''Yield mounted block devices that don't have entries in /etc/fstab'''
    for m in mounts():
        if not isblock(m.dev):
            continue
        if not any(m.mnt == d.mnt for d in mounts("/etc/fstab")):
            yield m

def loopmount(filename, mntpoint=None):
    if mntpoint is None:
        mntpoint = mkdtemp(prefix=__package__+'.mnt.')
    check_call(['mount', '-oloop', filename, mntpoint])
    for m in mounts():
        if m.mnt == mntpoint:
            return m

def fix_loop_entry(mnt, iso):
    '''return new FstabEntry with dev=backing_file and "loop" added to opts'''
    # losetup, due to the limitations of LOOP_GET_STATUS64 and the lack of backing_file
    # in sysfs, truncates the backing file output to < 64 characters. The backing file
    # must come from the caller.

    opts = ','.join([mnt.opts, "loop"])
    return mnt._replace(dev=realpath(iso), opts=opts)

def umount(mntpoint):
    try:
        check_call(['umount', '-d', mntpoint])
    except CalledProcessError:
        log.warn('umount %s failed: %s', mntpoint, e.output)
        log.warn('trying lazy umount')
        call(['umount', '-l', mntpoint])

# see systemd/src/shared/unit-name.c:do_escape()
validchars='0123456789'\
           'abcdefghijklmnopqrstuvwxyz'\
           'ABCDEFGHIJKLMNOPQRSTUVWXYZ'\
           ':-_.\\'

def systemd_escape_char(ch):
    if ch == '/':
        return '-'
    elif ch == '-' or ch == '\\' or ch not in validchars:
        return '\\x%x' % ord(ch)
    else:
        return ch

def systemd_escape(path):
    if path == '/':
        return '-'
    newpath = ''
    path = path.strip('/')
    if path[0] == '.':
        newpath += '\\x2e'
        path = path[1:]
    for ch in path:
        newpath += systemd_escape_char(ch)
    return newpath

def shell_escape(string):
    return "'%s'" % string.replace("'", "'\\''")

def write_prep_mount(mount, unitdir, iso, desc=None, unitopts=""):
    if desc is None:
        desc = "Upgrade Media"
    if isloop(mount.dev):
        mount = fix_loop_entry(mount, iso)
    unit = join(unitdir, systemd_escape(mount.mnt)+'.mount')
    with open(unit, 'w') as u:
        u.write("mount -t %s -o %s %s %s\n" %
                (shell_escape(mount.type),
                 shell_escape(mount.opts),
                 shell_escape(mount.dev),
                 shell_escape(mount.mnt)))
    os.chmod(unit, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
    return unit



#!/usr/bin/python
#
# redhat-upgrade-tool.py - commandline frontend for redhat-upgrade-tool, the
# Red Hat Upgrade Tool.
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import sys, time, platform, shutil, signal
from subprocess import CalledProcessError, Popen, PIPE
from ConfigParser import NoOptionError

from redhat_upgrade_tool.util import call, check_call, rm_f, mkdir_p, rlistdir
from redhat_upgrade_tool.download import UpgradeDownloader, YumBaseError, yum_plugin_for_exc, URLGrabError
from redhat_upgrade_tool.sysprep import prep_upgrade, prep_boot, setup_media_mount, setup_cleanup_post, disable_old_repos, Config
from redhat_upgrade_tool.sysprep import modify_repos, remove_cache, reset_boot
from redhat_upgrade_tool.boot import upgrade_boot_args
from redhat_upgrade_tool.upgrade import RPMUpgrade, TransactionError

from redhat_upgrade_tool.commandline import parse_args, do_cleanup, device_setup
from redhat_upgrade_tool import textoutput as output
from redhat_upgrade_tool import upgradeconf
from redhat_upgrade_tool import rhel_gpgkey_path
from redhat_upgrade_tool import preupgrade_script_path
from redhat_upgrade_tool import release_version_file
from redhat_upgrade_tool import _, kernelpath, initrdpath
from redhat_upgrade_tool import MIN_AVAIL_BYTES_FOR_BOOT

import redhat_upgrade_tool.logutils as logutils
import redhat_upgrade_tool.media as media

from preupg.xccdf import XccdfHelper
from preupg import settings

import logging
log = logging.getLogger("redhat-upgrade-tool")
def message(m):
    print m
    log.info(m)


def setup_downloader(version, instrepo=None, cacheonly=False, repos=[],
                     enable_plugins=[], disable_plugins=[], noverifyssl=False):
    log.debug("setup_downloader(version=%s, repos=%s)", version, repos)
    f = UpgradeDownloader(version=version, cacheonly=cacheonly)
    f.preconf.enabled_plugins += enable_plugins
    f.preconf.disabled_plugins += disable_plugins
    f.instrepoid = instrepo
    repo_cb = output.RepoCallback()
    repo_prog = output.RepoProgress(fo=sys.stderr)
    disabled_repos = f.setup_repos(callback=repo_cb,
                                   progressbar=repo_prog,
                                   repos=repos,
                                   noverifyssl=noverifyssl)
    disabled_repos = filter(lambda id: id != f.instrepoid, disabled_repos)
    if disabled_repos:
        print _("No upgrade available for the following repos") + ": " + \
                " ".join(disabled_repos)
        print _("Check that the repo URLs are correct.")
        log.info("disabled repos: " + " ".join(disabled_repos))
    return f

def download_packages(f):
    updates = f.build_update_transaction(callback=output.DepsolveCallback(f))
    # check for empty upgrade transaction
    if not updates:
        print _('No upgrade found, please check the repository specified is correct.')
        print _('Finished. Nothing to do.')
        raise SystemExit(0)
    # print dependency problems before we start the upgrade
    transprobs = f.describe_transaction_problems()
    if transprobs and not major_upgrade:
        print "WARNING: potential problems with upgrade"
        for p in transprobs:
            print "  " + p
    # clean out any unneeded packages from the cache
    f.clean_cache(keepfiles=(p.localPkg() for p in updates))
    # download packages
    f.download_packages(updates, callback=output.DownloadCallback())

    return updates

def transaction_test(pkgs):
    print _("testing upgrade transaction")
    pkgfiles = set(po.localPkg() for po in pkgs)
    fu = RPMUpgrade()
    probs = fu.setup_transaction(pkgfiles=pkgfiles, check_fatal=False)
    rv = fu.test_transaction(callback=output.TransactionCallback(numpkgs=len(pkgfiles)))
    return (probs, rv)

def reboot():
    call(['reboot'])

def get_preupgrade_result_name():
    return os.path.join(settings.assessment_results_dir,
                        settings.xml_result_name)

def check_release_version_file():
    if not os.path.exists(release_version_file):
        print _("File for checking if upgrade path is possible doesn't exist.")
        raise SystemExit(1)
    try:
        with open(release_version_file) as release_file:
            release = release_file.readlines()
        try:
            rel = release[1].strip()
            if rel != args.network:
                print _("You are trying to upgrade on system %s which is not allowed." % args.network)
                print _("Preupgrade-assistant assess the system for upgrade to %s version" % rel)
                raise SystemExit(1)
        except KeyError:
            print _("Release file doesn't contain proper versions.")
            raise SystemExit(1)
    except (IOError, OSError) as e:
        print _("Unable to read upgrade path file: %s") % e
        raise SystemExit(1)


def main(args):
    global major_upgrade

    if args.clean:
        do_cleanup(args)
        return
    else:
        # Leaving cache from previous runs of the tool could foil the correct
        # download of packages for upgrade (bz#1303982)
        remove_cache()

    if args.device or args.iso:
        device_setup(args)

    # Get our packages set up where we can use 'em
    print _("setting up repos...")
    f = setup_downloader(version=args.network,
                         cacheonly=args.cacheonly,
                         instrepo=args.instrepo,
                         repos=args.repos,
                         enable_plugins=args.enable_plugins,
                         disable_plugins=args.disable_plugins,
                         noverifyssl=args.noverifyssl)

    # Compare the first part of the version number in the treeinfo with the
    # first part of the version number of the system to determine if this is a
    # major version upgrade
    if not args.force and args.network:
        check_release_version_file()

    if f.treeinfo.get('general', 'version').split('.')[0] != \
            platform.linux_distribution()[1].split('.')[0]:

        major_upgrade = True

        # Check if preupgrade-assistant has been run
        if args.force:
            log.info("Skipping check for preupgrade-assisant")

        if not args.force:
            # Run preupg --riskcheck
            returncode = XccdfHelper.check_inplace_risk(get_preupgrade_result_name(), 0)
            if int(returncode) == 0:
                print _("Preupgrade assistant does not found any risks")
                print _("Upgrade will continue.")
            elif int(returncode) == 1:
                print _("Preupgrade assistant risk check found risks for this upgrade.")
                print _("You can run preupg --riskcheck --verbose to view these risks.")
                print _("Addressing high risk issues is required before the in-place upgrade")
                print _("and ignoring these risks may result in a broken upgrade and unsupported upgrade.")
                print _("Please backup your data.")
                print ""
                print _("List of issues:")

                XccdfHelper.check_inplace_risk(get_preupgrade_result_name(), verbose=2)

                # Python 2.6 raises EOFError if raw_input receives a SIGWINCH.
                # Try to tell the difference between that and a real EOF.

                global sigwinch
                sigwinch = False
                def handle_sigwinch(signum, frame):
                    global sigwinch
                    sigwinch = True
                orig_handler = signal.signal(signal.SIGWINCH, handle_sigwinch)

                while True:
                    try:
                        sigwinch = False
                        answer = raw_input(_("Continue with the upgrade [Y/N]? "))
                        break
                    except EOFError:
                        if sigwinch:
                            # Not a real EOF, try again
                            print
                            continue
                        else:
                            # Real EOF, exit
                            answer = ''
                            break
                signal.signal(signal.SIGWINCH, orig_handler)

                # TRANSLATORS: y for yes
                if answer.lower() != _('y'):
                    raise SystemExit(1)
            elif int(returncode) == 2:
                print _("preupgrade-assistant risk check found EXTREME risks for this upgrade.")
                print _("Run preupg --riskcheck --verbose to view these risks.")
                print _("Continuing with this upgrade is not recommended.")
                raise SystemExit(1)
            else:
                print _("preupgrade-assistant has not been run.")
                print _("To perform this upgrade, either run preupg or run redhat-upgrade-tool --force")
                raise SystemExit(1)

    # Check that we are upgrading to the same variant
    if not args.force:
        distro = platform.linux_distribution()[0]
        if not distro.startswith("Red Hat Enterprise Linux "):
            print _("Invalid distribution: %s") % distro
            raise SystemExit(1)

        from_variant = distro[len('Red Hat Enterprise Linux '):]
        try:
            to_variant = f.treeinfo.get('general', 'variant')
        except NoOptionError:
            print _("Upgrade repository is not a Red Hat Enterprise Linux repository")
            raise SystemExit(1)

        if from_variant != to_variant:
            print _("Upgrade requested from Red Hat Enterprise Linux %s to %s") % (from_variant, to_variant)
            print _("Upgrades between Red Hat Enterprise Linux variants is not supported.")
            raise SystemExit(1)
    else:
        log.info("Skipping variant check")

    if args.nogpgcheck:
        f._override_sigchecks = True
    elif not f.instrepo.gpgcheck:
        # If instrepo is a Red Hat repo, add the gpg key and reload the repos
        try:
            key = "product"
            if not f.treeinfo.has_section(key):
                key = "release"
            if f.treeinfo.get(key, 'name') == 'Red Hat Enterprise Linux':
                log.info("Reloading repos with GPG key")
                args.repos.append(('gpgkey', '%s=%s' % (f.instrepo.name, rhel_gpgkey_path)))
                f = setup_downloader(version=args.network,
                                     cacheonly=args.cacheonly,
                                     instrepo=args.instrepo,
                                     repos=args.repos,
                                     enable_plugins=args.enable_plugins,
                                     disable_plugins=args.disable_plugins,
                                     noverifyssl=args.noverifyssl)
        except NoOptionError:
            log.debug("No product name found, skipping gpg check")

    if args.expire_cache:
        print "expiring cache files"
        f.cleanExpireCache()
        return
    if args.clean_metadata:
        print "cleaning metadata"
        f.cleanMetadata()
        return

    # Cleanup old conf files
    log.info("Clearing %s", upgradeconf)
    rm_f(upgradeconf)
    mkdir_p(os.path.dirname(upgradeconf))

    # TODO: error msg generation should be shared between CLI and GUI
    if args.skipkernel:
        message("skipping kernel/initrd download")
    elif f.instrepoid is None or f.instrepoid in f.disabled_repos:
        print _("Error: can't get boot images.")
        if args.instrepo:
            print _("The '%s' repo was rejected by yum as invalid.") % args.instrepo
            if args.iso:
                print _("The given ISO probably isn't an install DVD image.")
                media.umount(args.device.mnt)
            elif args.device:
                print _("The media doesn't contain a valid install DVD image.")
        else:
            print _("The installation repo isn't currently available.")
            print _("Try again later, or specify a repo using --instrepo.")
        raise SystemExit(1)
    else:
        print _("getting boot images...")
        kernel, initrd = f.download_boot_images() # TODO: force arch?

    if args.skippkgs:
        message("skipping package download")
    else:
        print _("setting up update...")
        if len(f.pkgSack) == 0:
            print("no updates available in configured repos!")
            raise SystemExit(1)
        pkgs = download_packages(f)
        # Run a test transaction
        probs, rv = transaction_test(pkgs)


    # And prepare for upgrade
    # TODO: use polkit to get root privs for these things
    print _("setting up system for upgrade")
    if not args.skippkgs:
        prep_upgrade(pkgs)

    # Disable the RHEL-6 repos
    disable_old_repos()

    # Save the repo configuration
    f.save_repo_configs()

    #dump all configuration to upgrade.conf, other tools need to know
    #TODO:some items are structured, would be nice to unpack them
    with Config(upgradeconf) as conf:
        argsdict = args.__dict__
        for arg in argsdict:
            conf.set("config", arg.__str__(), argsdict[arg].__str__())

    if args.cleanup_post:
        setup_cleanup_post()

    # Workaround the redhat-upgrade-dracut upgrade-post hook order problem
    # Copy upgrade.conf to /root/preupgrade so that it won't be removed
    # before the postupgrade scripts are run.
    mkdir_p('/root/preupgrade')
    shutil.copyfile(upgradeconf, '/root/preupgrade/upgrade.conf')


    # Run the preuprade scripts if present
    if os.path.isdir(preupgrade_script_path):
        scripts = sorted(rlistdir(preupgrade_script_path))
        failed_scripts = {}
        for s in scripts:
            if os.access(s, os.X_OK):
                try:
                    check_call(s)
                except CalledProcessError as e:
                    failed_scripts[s] = e.returncode
        if failed_scripts:
            print("Following preupgrade script(s) failed:\n")
            for key, val in failed_scripts.iteritems():
                print("%s exited with status %d" % (key, val))
            print('exiting')
            sys.exit(1)

    if not args.skipbootloader:
        if args.skipkernel:
            print "warning: --skipkernel without --skipbootloader"
            print "using default paths: %s %s" % (kernelpath, initrdpath)
            kernel = kernelpath
            initrd = initrdpath
        upgrade_boot_args()
        prep_boot(kernel, initrd)

    # Check for available space in /boot/ needed for kernel and grub
    # installation during the upgrade.
    statvfs = os.statvfs("/boot")
    avail_bytes = statvfs.f_frsize * statvfs.f_bavail
    if avail_bytes < MIN_AVAIL_BYTES_FOR_BOOT:
        reset_boot()
        additional_mib_needed = \
            (MIN_AVAIL_BYTES_FOR_BOOT - avail_bytes) / 2**20
        sys.stderr.write(_("Not enough space. /boot/ needs additional %d MiB"
                           ".\n") % additional_mib_needed)
        raise SystemExit(1)

    # Replace temporary media paths
    modify_repos(args)

    if args.device:
        setup_media_mount(args.device, args.iso)

    if args.iso:
        media.umount(args.device.mnt)

    if args.reboot:
        reboot()
    else:
        print _('Finished. Reboot to start upgrade.')

    # --- Here's where we summarize potential problems. ---

    # list packages without updates, if any
    missing = sorted(f.find_packages_without_updates(), key=lambda p:p.nevra)
    if missing and not major_upgrade:
        message(_('Packages without updates:'))
        for p in missing:
            message("  %s" % p)

    # warn if the "important" repos are disabled
    #if f.disabled_repos:
        # NOTE: I hate having a hardcoded list of Important Repos here.
        # This information should be provided by the system, somehow..
        #important = ("fedora", "updates")
        #if any(i in f.disabled_repos for i in important):
        #    msg = _("WARNING: Some important repos could not be contacted: %s")
        #else:
        #    msg = _("NOTE: Some repos could not be contacted: %s")
        #print msg % ", ".join(f.disabled_repos)
        #print _("If you start the upgrade now, packages from these repos will not be installed.")

    # warn about broken dependencies etc.
    # If this is a major version upgrade, the user has already been warned
    # about all of this from preupgrade-assistant, so skip the warning here
    if probs and not major_upgrade:
        print
        print _("WARNING: problems were encountered during transaction test:")
        for s in probs.summaries:
            print "  "+s.desc
            for line in s.format_details():
                print "    "+line
        print _("Continue with the upgrade at your own risk.")

if __name__ == '__main__':
    args = parse_args()
    major_upgrade = False

    # TODO: use polkit to get privs for modifying bootloader stuff instead
    if os.getuid() != 0:
        print _("you must be root to run this program.")
        raise SystemExit(1)

    # set up logging
    if args.debuglog:
        logutils.debuglog(args.debuglog)
    logutils.consolelog(level=args.loglevel)
    log.info("%s starting at %s", sys.argv[0], time.asctime())

    try:
        exittype = "cleanly"
        main(args)
    except KeyboardInterrupt as e:
        print
        log.info("exiting on keyboard interrupt")
        message(_("Exiting on keyboard interrupt"))
        raise SystemExit(1)
    except (YumBaseError, URLGrabError) as e:
        print
        if hasattr(e, "value") and isinstance(e.value, list):
            err = e.value.pop(0)
            message(_("Downloading failed: %s") % err)
            for p in e.value:
                message("  %s" % p)
        else:
            message(_("Downloading failed: %s") % e)
        log.debug("Traceback (for debugging purposes):", exc_info=True)
        raise SystemExit(2)
    except TransactionError as e:
        print
        message(_("Upgrade test failed with the following problems:"))
        for s in e.summaries:
            message(s)
        log.debug("Detailed transaction problems:")
        for p in e.problems:
            log.debug(p)
        log.error(_("Upgrade test failed."))
        raise SystemExit(3)
    except Exception as e:
        pluginfile = yum_plugin_for_exc()
        if pluginfile:
            plugin, ext = os.path.splitext(os.path.basename(pluginfile))
            log.error(_("The '%s' yum plugin has crashed.") % plugin)
            log.error(_("Please report this problem to the plugin developers:"),
                      exc_info=True)
            raise SystemExit(1)
        log.info("Exception:", exc_info=True)
        exittype = "with unhandled exception"
        raise
    finally:
        log.info("%s exiting %s at %s", sys.argv[0], exittype, time.asctime())



# boot - bootloader config modification code
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

from .util import check_output, check_call, PIPE, Popen, CalledProcessError
from shutil import copyfileobj

import logging
log = logging.getLogger(__package__+".boot")

kernelprefix = "/boot/vmlinuz-"

def kernelver(kernel):
    if kernel.startswith(kernelprefix):
        return kernel.split(kernelprefix,1)[1]
    else:
        raise ValueError("kernel name must start with '%s'" % kernelprefix)

def add_entry(kernel, initrd, banner=None, kargs=[], makedefault=True, remove_kargs=[]):
    cmd = ["new-kernel-pkg", "--initrdfile", initrd]
    if banner:
        cmd += ["--banner", banner]
    if kargs:
        cmd += ["--kernel-args", " ".join(kargs)]
    if makedefault:
        cmd += ["--make-default"]
    cmd += ["--install", kernelver(kernel)]
    output = check_output(cmd)

    if remove_kargs:
        # Update the entry to remove arguments pulled in from the default entry
        cmd = ["new-kernel-pkg", "--remove-args", " ".join(remove_kargs),
                "--update", kernelver(kernel)]
        output += check_output(cmd)

    return output

def remove_entry(kernel):
    cmd = ["new-kernel-pkg", "--remove", kernelver(kernel)]
    return check_output(cmd)

def initramfs_append_files(initramfs, files):
    '''Append the given files to the named initramfs.
       Raises IOError if the files can't be read/written.
       Raises CalledProcessError if cpio returns a non-zero exit code.'''
    if isinstance(files, basestring):
        files = [files]
    filelist = ''.join(f+'\n' for f in files if open(f))
    with open(initramfs, 'ab') as outfd:
        cmd = ["cpio", "-co"]
        cpio = Popen(cmd, stdin=PIPE, stdout=outfd, stderr=PIPE)
        (out, err) = cpio.communicate(input=filelist)
        if cpio.returncode:
            raise CalledProcessError(cpio.returncode, cmd, err)

def initramfs_append_images(initramfs, images):
    '''Append the given images to the named initramfs.
       Raises IOError if the files can't be read/written.'''
    with open(initramfs, 'ab') as outfd:
        for i in images:
            with open(i, 'rb') as infd:
                copyfileobj(infd, outfd)

def need_mdadmconf():
    '''Does this system need /etc/mdadm.conf to boot?'''
    # NOTE: there are probably systems that have mdadm.conf but don't require
    # it to boot, but I don't know how you tell the difference, so...
    try:
        for line in open("/etc/mdadm.conf"):
            line = line.strip()
            if line and not line.startswith("#"):
                # Hey there's actual *data* in here! WE MIGHT NEED THIS!!
                return True
    except IOError:
        pass
    return False

def upgrade_boot_args():
    '''function checks if all boot parameters are fine

       This function will modify the arguments for the current default bootloader
       entry. dracut in RHEL 6 can accept either the old or new bootloader
       arguments, but systemd in RHEL 7 requires the new arguments. This
       function must be called to modify the current bootloader entry, which
       will be used as the template for both the System Upgrade entry and the
       entry created during the kernel upgrade.

       This change is not undone by --resetbootloader.
    '''
    replaced_options = {'rdbreak': 'rd.break', 'rd_DASD_MOD': 'rd.dasd',
                        'rdinitdebug': 'rd.debug', 'rdnetdebug': 'rd.debug',
                        'rdblacklist': 'rd.driver.blacklist', 'rdinsmodpost': 'rd.driver.post',
                        'rdloaddriver': 'rd.driver.pre', 'rdinfo': 'rd.info', 'check': 'rd.live.check',
                        'rdlivedebug': 'rd.live.debug', 'live_dir': 'rd.live.dir',
                        'liveimg': 'rd.live.image', 'overlay': 'rd.live.overlay',
                        'readonly_overlay': 'rd.live.overlay.readonly', 'reset_overlay': 'rd.live.overlay.reset',
                        'live_ram': 'rd.live.ram', 'rdshell': 'rd.shell', 'rd_NO_SPLASH': 'rd.splash',
                        'rdudevdebug': 'rd.udev.debug', 'rdudevinfo': 'rd.udev.info',
                        'KEYMAP': 'vconsole.keymap', 'KEYTABLE': 'vconsole.keymap',
                        'SYSFONT': 'vconsole.font', 'CONTRANS': 'vconsole.font.map',
                        'UNIMAP': 'vconsole.font.unimap', 'UNICODE': 'vconsole.unicode',
                        'EXT_KEYMAP': 'vconsole.keymap.ext', 'LANG': 'rd.locale.LANG'}
    no_options = {'rd_NO_DM': 'rd.dm=0', 'rd_NO_DM': 'rd.dm=0', 'rd_NO_LVM': 'rd.lvm=0',
                  'rd_NO_MD': 'rd.md=0', 'rd_NO_LUKS': 'rd.luks=0', 'rd_NO_CRYPTTAB': 'rd.luks.crypttab=0',
                  'rd_NO_PLYMOUTH': 'rd.plymouth=0', 'rd_NO_MDADMCONF': 'rd.md.conf=0',
                  'rd_NO_LVMCONF': 'rd.lvm.conf', 'rd_NO_MDIMSM': 'rd.md.imsm=0', 'rd_NO_MULTIPATH': 'rd.multipath=0',
                  'rd_NO_ZFCPCONF': 'rd.zfcp.conf=0', 'rd_NO_FSTAB': 'rd.fstab=0',
                  'iscsi_firmware': 'rd.iscsi.firmware=0'}
    translate_options = ['rd_NFS_DOMAIN', 'rd_LVM_SNAPHOST', 'rd_LVM_SNAPSIZE', 'rd_LVM_VG',
                         'rd_LUKS_KEYPATH', 'rd_LUKS_UUID', 'rd_LVM_LV', 'rd_retry',
                         'rd_ZNET', 'rd_ZFCP', 'rd_CCW', 'rd_DM_UUID', 'rd_MD_UUID',
                         'rd_LUKS_KEYDEV_UUID',
                         ]
    iscsi_options = ['iscsi_initiator', 'iscsi_target_name', 'iscsi_target_ip', 'iscsi_target_port',
                     'iscsi_target_group', 'iscsi_username',
                     'iscsi_password', 'iscsi_in_username', 'iscsi_in_password']

    orig_args = []
    new_args = []

    # Use grubby to get the current list of arguments
    kernel = check_output(['grubby', '--default-kernel']).strip()
    kinfo = check_output(['grubby', '--info=%s' % kernel]).split('\n')

    # Look for the line starting with args=, remove the args= and the quotes,
    # split the line into a list.
    for line in kinfo:
        if line.startswith('args='):
            orig_args = line[6:-1].split()
            break

    for arg in orig_args:
        arg_name = arg.split('=')[0]

        if arg_name in replaced_options:
            arg = arg.replace(arg_name, replaced_options[arg_name])

        if arg_name in no_options:
            arg = arg.replace(arg_name, no_options[arg_name])

        if arg_name in translate_options:
            arg = arg.replace(arg_name, arg_name.replace('_', '.').lower())

        if arg_name in iscsi_options:
            arg = 'rd.' + arg.replace(arg_name, '_', '.')

        new_args.append(arg)

    # Use new-kernel-pkg to remove the old list of arguments and add back the new one
    log.info("Upgrading kernel args for %s", kernel)
    log.debug("Old args: %s", orig_args)
    log.debug("New args: %s", new_args)
    check_call(['new-kernel-pkg', '--remove-args', ' '.join(orig_args),
        '--update', kernelver(kernel)])
    check_call(['new-kernel-pkg', '--kernel-args', ' '.join(new_args),
        '--update', kernelver(kernel)])



# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>
'''
Configuration file parser / writer for .treeinfo files.

The .treeinfo file lives in the top directory of an installable distribution
'tree' and describes the contents - boot images/media, packages, etc.

The file syntax is the normal ConfigParser .ini-style format.

A typical .treeinfo file looks like this:

    [general]
    family = Red Hat Enterprise Linux
    timestamp = 1337720130.41
    variant = Server
    version = 7.0
    packagedir =
    arch = x86_64

    [stage2]
    mainimage = LiveOS/squashfs.img

    [images-x86_64]
    kernel = images/pxeboot/vmlinuz
    initrd = images/pxeboot/initrd.img
    boot.iso = images/boot.iso

    [images-xen]
    kernel = images/pxeboot/vmlinuz
    initrd = images/pxeboot/initrd.img

    [checksums]
    images/boot.iso = sha256:[long hex string]
    images/macboot.img = sha256:[long hex string]
    images/efiboot.img = sha256:[long hex string]
    images/pxeboot/initrd.img = sha256:[long hex string]
    images/pxeboot/vmlinuz = sha256:[long hex string]
    repodata/repomd.xml = sha256:[long hex string]

Some details about the file format:

* [general] section: general info about this distro tree.

  - 'timestamp', 'arch', and 'version' are required.

  - 'timestamp' is the build time, in seconds since the Epoch.
     It may be a floating point number.

  - 'timestamp' should match the timestamp in other .treeinfo files for other
     arches/variants built as part of the same "compose".
     (e.g. RHEL 7.0 x86_64 and RHEL 7.0 i386 have the same timestamp.)

  - 'version' is a string, not an integer - "17-Beta" is valid.
  - 'family' is the OS/distro family name, 'variant' is name of the distro
    variant, like "Server" or "Client" or "Workstation".

  - 'packagedir' usually points to the directory containing packages, but in
     RHEL this is ignored in favor of reading repodata/repomd.xml.

  - 'discnum' and 'totaldiscs' can be used when writing a tree to multiple
     CD/DVD images.


* [stage2] section: installer runtime images.

  - 'mainimage' will refer to the main installer runtime image, if one exists.
     In RHEL 7.0 and later the installer uses this to automatically find
     its runtime if you boot with 'inst.repo=url://to/os/tree'.


* [images-*] sections: list the boot images (kernel, initramfs, boot.iso, etc.)

  - There should be an [images-$arch] section for the arch in [general].
    There may be more sections for other major arches the tree supports.

  - [images-xen], by convention, contains the images that should be used when
    installing into a virtualized guest.

  - Standard keys are 'kernel', 'initrd', and 'boot.iso'.


* [checksums] section: checksums of various files in the tree

  - Everything listed in all the [images-*] sections should have a checksum.
  - Each line is of the form:
    relative/path = hash_algorithm:hex_digest

'''

import ConfigParser
from ConfigParser import RawConfigParser
import hashlib
import time
from os.path import join, normpath
import logging
from StringIO import StringIO

# TODO: release this separately so it can be used by other stuff
#       (pungi, libvirt, etc.)
#log = logging.getLogger('treeinfo')
#log.addHandler(logging.NullHandler())
log = logging.getLogger(__package__+".treeinfo")

def hexdigest(filename, algo, blocksize=8192):
    hasher = hashlib.new(algo)
    with open(filename, 'rb') as fobj:
        while True:
            data = fobj.read(blocksize)
            if not data:
                break
            hasher.update(data)
    return hasher.hexdigest()

__all__ = ['Treeinfo', 'TreeinfoError']

# Base class for Treeinfo errors.
TreeinfoError = ConfigParser.Error

class Treeinfo(RawConfigParser):
    '''
    A subclass of RawConfigParser with some extra bits for handling .treeinfo
    files, such as are written by pungi and friends.
    '''
    def __init__(self, fromfile=None, topdir=None):
        '''
        fromfile can be a file-like object (anything with a .readline method)
        or a filename, or a list of filenames.

        topdir specifies the default topdir that any 'relpath' arguments are
        assumed to be relative to (see add_image, add_checksum, etc.)
        '''
        try:
            RawConfigParser.__init__(self, allow_no_value=True)
        except TypeError: # python 2.6 :/
            RawConfigParser.__init__(self)
        self._fullpath = dict() # save relpath -> filename mappings here
        if hasattr(fromfile, 'readline'):
            self.readfp(fromfile)
        elif fromfile is not None:
            self.read(fromfile)
        self.topdir = topdir

    def _path(self, relpath, topdir=None):
        if relpath not in self._fullpath:
            if topdir is None:
                topdir = self.topdir or '.'
            self._fullpath[relpath] = normpath(join(topdir, relpath))
        return self._fullpath[relpath]

    def read_str(self, data):
        self.readfp(StringIO(data)) # guh

    def setopt(self, section, option, value):
        '''Sets the given option, creating the section if it doesn't exist'''
        if not self.has_section(section):
            self.add_section(section)
        self.set(section, option, value)

    def get_image(self, arch, imgtype):
        '''return the relative path to the image type for the given arch.
        typical imgtypes are: kernel, initrd, boot.iso'''
        return self.get('images-%s' % arch, imgtype)

    def image_arches(self):
        '''return a list of arches mentioned in 'images-*' sections'''
        return [s[7:] for s in self.sections() if s.startswith('images-')]

    def checkvalues(self):
        '''Check the .treeinfo to make sure it has all required elements.'''
        for f in ('version', 'arch'):
            self.get('general', f)
        # TODO check for checksums for all images

    def checkfile(self, filename, relpath):
        '''
        Check the given file against the info in [checksum].

        filename must be a full path to the file to be checked.
        relpath is the relative path that was used to fetch the file,
        i.e. the value from the [images-*] section (and the key in the
        [checksums] section)
        '''
        val = self.get('checksums', relpath)
        algo, checksum = val.split(':',1)
        try:
            return checksum == hexdigest(filename, algo)
        except IOError:
            return False

    def add_image(self, arch, imgtype, relpath, topdir=None, algo='sha256'):
        '''
        Add an image to the .treeinfo file: adds an entry to the [images-$arch]
        section with imgtype as the key and relpath as the value.

        If algo is not none, also checksums the file and adds a corresponding
        entry to the [checksums] section, of the form:
          relpath = algo:checksum_in_hex

        If topdir is None, the Treeinfo.topdir value is used.
        '''
        log.debug("adding %s %s: %s", arch, imgtype, relpath)
        section = 'images-%s' % arch
        self.setopt(section, imgtype, relpath)
        if algo:
            self.add_checksum(relpath, topdir, algo)

    def add_checksum(self, relpath, topdir=None, algo='sha256'):
        '''
        Add an item to the [checksums] section for the given file.

        relpath is the path to the file, relative to .treeinfo.
        topdir is the directory that filename is relative to.
        algo is the checksum algorithm.
        '''
        fullpath = self._path(relpath, topdir)
        log.debug("add_checksum(%s)", fullpath)
        self.setopt('checksums', relpath, algo+':'+hexdigest(fullpath, algo))
        log.debug("%s = %s" % (relpath, self.get('checksums',relpath)))

    def add_timestamp(self, timestamp=None):
        '''
        Add a 'timestamp' entry to the [general] section.
        Uses the current time if timestamp is not set.
        '''
        self.setopt('general', 'timestamp', timestamp or time.time())

    def writetreeinfo(self, topdir=None, strict=True, add_timestamp=False):
        if not (topdir or self.topdir):
            raise TypeError("writetreeinfo() requires topdir")
        if strict:
            self.checkvals()
        if add_timestamp:
            self.add_timestamp()
        # TODO sort checksums to end
        with open(self._path('.treeinfo'), 'w') as fp:
            RawConfigParser.write(fp)


# TODO: unit tests



'''terminal.py - provides info about the size of the controlling terminal.

Set ttyfd to choose which fd to use as the controlling tty (default: stdout)

NOTE: Importing this module adds a handler for SIGWINCH so it can update 'size'
whenever the controlling tty changes size.'''

import fcntl, struct, termios, signal, warnings
from collections import namedtuple

size = None
ttyfd = 1

class winsize(namedtuple('winsize', 'rows cols')):
    '''
    The current size of the terminal.

    rows: Number of rows on the controlling terminal
    cols: Number of columns on the controlling terminal
    '''
    pass

def getsize(fd=ttyfd):
    '''Return the size of the tty attached to the given fd (default: stdin)'''
    try:
        buf = fcntl.ioctl(fd, termios.TIOCGWINSZ, struct.pack('8x'))
        (rows, cols, _, _) = struct.unpack('hhhh', buf)
        size = winsize(rows, cols)
    except IOError:
        size = winsize(0, 0)
    return size

size = getsize()

# If we're being reloaded, save a reference to the old signal handler
try:
    old_handle_winch = handle_winch
except NameError:
    old_handle_winch = signal.SIG_DFL

# Update 'size' when we get SIGWINCH
def handle_winch(signum, frame):
    global size
    size = getsize()
old_handler = signal.signal(signal.SIGWINCH, handle_winch)

# Warn the user if we just clobbered an external SIGWINCH handler
if old_handler not in (signal.SIG_DFL, signal.SIG_IGN, old_handle_winch):
    if old_handler is None:
        msg = "clobbered existing non-Python SIGWINCH handler"
    else:
        msg = "clobbered existing SIGWINCH handler %s" % old_handler
    warnings.warn(msg, RuntimeWarning)



# logutils.py - logging utility functions
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import logging

class CompatNullHandler(logging.Handler):
    def emit(self, record):
        pass

try:
    from logging import NullHandler
except ImportError:
    NullHandler = CompatNullHandler

class Formatter(logging.Formatter):
    levelsyms = {
        logging.DEBUG:   '(DD)',
        logging.INFO:    '(II)',
        logging.WARNING: '(WW)',
        logging.ERROR:   '(EE)',
        logging.CRITICAL:'(CC)',
        logging.FATAL:   '(FF)',
    }

    defaultfmt="[%(reltime)10.3f] %(levelsym)s %(name)s:%(funcName)s() %(message)s"
    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = self.defaultfmt
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        record.reltime = float(record.relativeCreated / 1000)
        record.levelsym = self.levelsyms.get(record.levelno, '(--)')
        if record.levelno < logging.DEBUG:
            record.levelsym = '(D%d)' % (10-record.levelno)
        return logging.Formatter.format(self, record)

def debuglog(filename, level=logging.DEBUG, loggername=__package__):
    # for even more debugging info:
    #import yum.logginglevels
    #level=yum.logginglevels.DEBUG_2
    h = logging.FileHandler(filename)
    h.setLevel(level)
    h.setFormatter(Formatter())
    logger = logging.getLogger(loggername)
    logger.setLevel(level)
    logger.addHandler(h)

def consolelog(level=logging.WARNING, loggername=__package__, tty=None):
    h = logging.StreamHandler(tty)
    h.setLevel(level)
    formatter = logging.Formatter('%(name)s %(levelname)s: %(message)s')
    h.setFormatter(formatter)
    logger = logging.getLogger(loggername)
    if level < logger.getEffectiveLevel():
        logger.setLevel(level)
    logger.addHandler(h)



# conf.py - config parser
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os
from tempfile import mkstemp
from ConfigParser import *

class Config(RawConfigParser):
    def __init__(self, filename, defaults=None):
        RawConfigParser.__init__(self, defaults)
        self.filename = filename
        self.read(filename)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.write()

    def writefp(self, fp):
        RawConfigParser.write(self, fp)

    def write(self):
        with open(self.filename, 'w') as outf:
            self.writefp(outf)

    def add_section(self, section, duplicate_ok=True):
        '''Add the named section, raising DuplicateSectionError if the section
        exists and duplicate_ok is False.'''
        try:
            RawConfigParser.add_section(self, section)
        except DuplicateSectionError:
            if not duplicate_ok:
                raise

    def set(self, section, option, value=None):
        '''Set an option, creating the section if needed.'''
        self.add_section(section)
        RawConfigParser.set(self, section, option, value)

    def get(self, section, option):
        '''Get an option, returning None if missing'''
        value = None
        try:
            value = RawConfigParser.get(self, section, option)
        except (NoSectionError, NoOptionError):
            pass
        return value



# __init__.py for the system upgrade python package
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

'''
NOTE WELL, DEAR READER: None of this is intended as a public API!  I reserve
the right to change anything and everything about how this library works at the
merest whim.

If you're actually trying to write something that uses this library, please
contact me (or whatever people/constructs are maintaining this after I quit
computers forever) to discuss how we could best design a sensible API that
would actually work for you.

Your pal,
-w
'''

import logging
import os
from .logutils import NullHandler
log = logging.getLogger(__package__)
log.addHandler(NullHandler())

import gettext
t = gettext.translation(__package__, "/usr/share/locale", fallback=True)
_ = t.lgettext

kernel_id = "redhat-upgrade-tool"
# NOTE: new-kernel-pkg requires this kernel name/path
kernelpath = '/boot/vmlinuz-%s' % kernel_id
initrdpath = '/boot/initramfs-%s.img' % kernel_id

cachedir = '/var/tmp/system-upgrade'
packagedir = '/var/lib/system-upgrade'
packagelist = os.path.join(packagedir, 'package.list')
upgradeconf = os.path.join(packagedir, 'upgrade.conf')
upgradelink = '/system-upgrade'
upgraderoot = '/system-upgrade-root'

mirrormanager = ''
defaultkey = ''

update_img_dir = os.path.join('/etc', __package__, 'update.img.d')

# FIXME: there is probably a better way to do this than hardcoding it
# This key id is used with RHEL-7.0 GA
rhel_gpgkey_id = 'fd431d51-4ae0493b'
rhel_gpgkey_path = 'file:///etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-release'

# Path for pre-upgrade scripts from preupgrade
preupgrade_dir = "/root/preupgrade"
preupgrade_script_path = os.path.join(preupgrade_dir, 'preupgrade-scripts')
release_version_file = os.path.join(preupgrade_dir, preupgrade_script_path, 'release_version')
grub_conf_file = "/boot/grub/grub.conf"

MIN_AVAIL_BYTES_FOR_BOOT = 50 * 2**20  # 50 MiB



# commandline.py - commandline parsing functions
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os, optparse, platform, sys
from copy import copy

from . import media
from . import packagedir
from .sysprep import reset_boot, remove_boot, remove_cache, misc_cleanup
from . import _

import logging
log = logging.getLogger(__package__)

def parse_args(gui=False):
    p = optparse.OptionParser(option_class=Option,
        description=_('Prepare system for upgrade.'),
        # Translators: This is the CLI's "usage:" string
        usage=_('%(prog)s <SOURCE> [options]' % {'prog': os.path.basename(sys.argv[0])}),
    )

    # === basic options ===
    p.add_option('-v', '--verbose', action='store_const', dest='loglevel',
        const=logging.INFO, help=_('print more info'))
    p.add_option('-d', '--debug', action='store_const', dest='loglevel',
        const=logging.DEBUG, help=_('print lots of debugging info'))
    p.set_defaults(loglevel=logging.WARNING)

    p.add_option('-f', '--force', action='store_true', default=False,
            help=_('continue even if preupgrade-assistant risk check fails'))
    p.add_option('--cleanup-post', action='store_true', default=False,
            help=_('cleanup old package after the upgrade'))

    p.add_option('--debuglog', default='/var/log/%s.log' % __package__,
        help=_('write lots of debugging output to the given file'))

    p.add_option('--reboot', action='store_true', default=False,
        help=_('automatically reboot to start the upgrade when ready'))


    # === hidden options. FOR DEBUGGING ONLY. ===
    p.add_option('--skippkgs', action='store_true', default=False,
        help=optparse.SUPPRESS_HELP)
    p.add_option('--skipkernel', action='store_true', default=False,
        help=optparse.SUPPRESS_HELP)
    p.add_option('--skipbootloader', action='store_true', default=False,
        help=optparse.SUPPRESS_HELP)
    p.add_option('-C', '--cacheonly', action='store_true', default=False,
        help=optparse.SUPPRESS_HELP)


    # === yum options ===
    yumopts = p.add_option_group(_('yum options'))
    yumopts.add_option('--enableplugin', metavar='PLUGIN',
        action='append', dest='enable_plugins', default=[],
        help=_('enable yum plugins by name'))
    yumopts.add_option('--disableplugin', metavar='PLUGIN',
        action='append', dest='disable_plugins', default=[],
        help=_('disable yum plugins by name'))
    yumopts.add_option('--nogpgcheck', action='store_true', default=False,
        help=_('disable GPG signature checking'))


    # === <SOURCE> options ===
    req = p.add_option_group(_('options for <SOURCE>'),
                               _('Location to search for upgrade data.'))
    req.add_option('--device', metavar='DEV',
        type="device_or_mnt",
        help=_('device or mountpoint. default: check mounted devices'))
    req.add_option('--iso', type="isofile",
        help=_('installation image file'))
    # Translators: This is for '--network [VERSION]' in --help output
    req.add_option('--network', metavar=_('VERSION'), type="VERSION",
        help=_('online repos matching VERSION (a number or "rawhide")'))


    # === options for --network ===
    net = p.add_option_group(_('additional options for --network'))
    net.add_option('--enablerepo', metavar='REPOID', action='callback', callback=repoaction,
        dest='repos', type=str, help=_('enable one or more repos (wildcards allowed)'))
    net.add_option('--disablerepo', metavar='REPOID', action='callback', callback=repoaction,
        dest='repos', type=str, help=_('disable one or more repos (wildcards allowed)'))
    net.add_option('--addrepo', metavar='REPOID=[@]URL',
        action='callback', callback=repoaction, dest='repos', type=str,
        help=_('add the repo at URL (@URL for mirrorlist)'))
    net.add_option('--instrepo', metavar='REPOID', type=str,
        help=_('get upgrader boot images from REPOID (default: auto)'))
    net.add_option('--instrepokey', metavar='GPGKEY', type='gpgkeyfile',
        help=_('use this GPG key to verify upgrader boot images'))
    net.add_option('--noverifyssl', action='store_true', default=False,
        help=_('do not verify the SSL certificate for HTTPS connections'))
    p.set_defaults(repos=[])

    if not gui:
        clean = p.add_option_group(_('cleanup commands'))

        clean.add_option('--resetbootloader', action='store_const',
            dest='clean', const='bootloader', default=None,
            help=_('remove any modifications made to bootloader'))
        clean.add_option('--clean', action='store_const', const='all',
            help=_('clean up everything written by %s') % __package__)
        p.add_option('--expire-cache', action='store_true', default=False,
            help=optparse.SUPPRESS_HELP)
        p.add_option('--clean-metadata', action='store_true', default=False,
            help=optparse.SUPPRESS_HELP)

    args, _leftover = p.parse_args()

    if not (gui or args.network or args.device or args.iso or args.clean):
        p.error(_('SOURCE is required (--network, --device, --iso)'))

    # allow --instrepo URL as shorthand for --repourl REPO=URL --instrepo REPO
    if args.instrepo and '://' in args.instrepo:
        args.repos.append(('add', 'cmdline-instrepo=%s' % args.instrepo))
        args.instrepo = 'cmdline-instrepo'

    # If network is requested, require an instrepo
    if args.network and not args.instrepo:
        p.error(_('--instrepo is required with --network'))

    if args.instrepo and args.instrepokey:
        args.repos.append(('gpgkey', 'instrepo=%s' % args.instrepokey))

    if not gui:
        if args.clean:
            args.resetbootloader = True

    return args

def repoaction(option, opt_str, value, parser, *args, **kwargs):
    '''Hold a list of repo actions so we can apply them in the order given.'''
    action = ''
    if opt_str.startswith('--enable'):
        action = 'enable'
    elif opt_str.startswith('--disable'):
        action = 'disable'
    elif opt_str.startswith('--addrepo'):
        action = 'add'
        # validate the argument
        repoid, eq, url = value.partition("=")
        if not (repoid and eq and "://" in url):
            raise optparse.OptionValueError(_("value should be REPOID=[@]URL"))
    parser.values.repos.append((action, value))

# check the argument to '--device' to see if it refers to install media
def device_or_mnt(option, opt, value):
    # Handle the default for --device=''
    if not value:
        value = 'auto'

    if value == 'auto':
        localmedia = media.find()
    else:
        # Canonicalize the device or mountpoint argument
        value = os.path.realpath(value)

        localmedia = [m for m in media.find() if value in (m.dev, m.mnt)]

    if len(localmedia) == 1:
        return localmedia.pop()

    if not localmedia:
        msg = _("no install media found - please mount install media first")
        if value != 'auto':
            msg = "%s: %s" % (value, msg)
    else:
        devs = ", ".join(m.dev for m in localmedia)
        msg = _("multiple devices found. please choose one of (%s)") % devs
    raise optparse.OptionValueError(msg)

# check the argument to '--iso' to make sure it's somewhere we can use it
def isofile(option, opt, value):
    if not os.path.exists(value):
        raise optparse.OptionValueError(_("File not found: %s") % value)
    if not os.path.isfile(value):
        raise optparse.OptionValueError(_("Not a regular file: %s") % value)
    if not media.isiso(value):
        raise optparse.OptionValueError(_("Not an ISO 9660 image: %s") % value)
    if any(value.startswith(d.mnt) for d in media.removable()):
        raise optparse.OptionValueError(_("ISO image on removable media\n"
            "Sorry, but this isn't supported yet.\n"
            "Copy the image to your hard drive or burn it to a disk."))
    return value

# valiadate a GPGKEY argument and return a URI ('file:///...')
def gpgkeyfile(option, opt, value):
    if value.startswith('file://'):
        value = value[7:]
    gpghead = '-----BEGIN PGP PUBLIC KEY BLOCK-----'
    try:
        with open(value) as keyfile:
            keyhead = keyfile.read(len(gpghead))
    except (IOError, OSError) as e:
        raise optparse.OptionValueError(e.strerror)
    if keyhead != gpghead:
        raise optparse.OptionValueError(_("File is not a GPG key"))
    return 'file://' + os.path.abspath(arg)

def VERSION(option, opt, value):
    if value.lower() == 'rawhide':
        return 'rawhide'

    distro, version, id = platform.linux_distribution()
    version = float(version)

    try:
        value = float(value)
    except ValueError:
        # Check if the option was missing
        if value[0] == '-':
            msg = _("%s option requires an argument") % opt
        else:
            msg = _("Invalid value for version")
        raise optparse.OptionValueError(msg)

    if value >= version:
        return str(value)
    else:
        msg = _("version must be greater than %i") % version
        raise optparse.OptionValueError(msg)

class Option(optparse.Option):
    TYPES = optparse.Option.TYPES + ("device_or_mnt", "isofile", "VERSION", "gpgkeyfile")
    TYPE_CHECKER = copy(optparse.Option.TYPE_CHECKER)

    TYPE_CHECKER["device_or_mnt"] = device_or_mnt
    TYPE_CHECKER["isofile"] = isofile
    TYPE_CHECKER["VERSION"] = VERSION
    TYPE_CHECKER["gpgkeyfile"] = gpgkeyfile

def do_cleanup(args):
    # FIXME: This installs RHSM product id certificates in case that
    # redhat-upgrade-dracut have not installed them.
    # It may be dropped when new redhat-upgrade-dracut is part of
    # install images.
    try:
        for cert in filter(lambda fn: fn.endswith('.pem'), os.listdir(packagedir)):
            old_fn = os.path.join(packagedir, cert)
            new_fn = '/etc/pki/product/%s' % cert
            print "Installing product cert %s to %s" % (old_fn, new_fn)
            os.rename(old_fn, new_fn)
    except OSError as e:
        import errno
        if e.errno != errno.ENOENT:
            raise

    if not args.skipbootloader:
        print "resetting bootloader config"
        reset_boot()
    if args.clean == 'bootloader':
        return
    if not args.skipkernel:
        print "removing boot images"
        remove_boot()
    if not args.skippkgs:
        print "removing downloaded packages"
        remove_cache()
    print "removing miscellaneous files"
    misc_cleanup()

def device_setup(args):
    # treat --device like --addrepo REPO=file://$MOUNTPOINT
    if args.device:
        args.repos.append(('add', 'upgradedevice=file://%s' % args.device.mnt))
        args.instrepo = 'upgradedevice'
    elif args.iso:
        try:
            args.device = media.loopmount(args.iso)
        except media.CalledProcessError as e:
            log.info("mount failure: %s", e.output)
            message('--iso: '+_('Unable to open %s') % args.iso)
            raise SystemExit(2)
        else:
            args.repos.append(('add', 'upgradeiso=file://%s' % args.device.mnt))
            args.instrepo = 'upgradeiso'



# download.py - yum-based download/depsolver for system upgrades
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import yum
import struct
import logging
from .callback import BaseTsCallback
from .treeinfo import Treeinfo, TreeinfoError
from .conf import Config
from yum.Errors import YumBaseError
from yum.parser import varReplace
from yum.constants import TS_REMOVE_STATES
from urlgrabber.grabber import URLGrabError
from yum.misc import gpgme
from gzip import GzipFile
from rhsm.certificate import create_from_pem

enabled_plugins = ['blacklist', 'whiteout']
disabled_plugins = ['rpm-warm-cache', 'remove-with-leaves', 'presto',
                    'auto-update-debuginfo', 'refresh-packagekit']

from . import _
from . import cachedir, upgradeconf, kernelpath, initrdpath, defaultkey
from . import mirrormanager
from . import packagedir
from .util import listdir, mkdir_p, rm_rf
from shutil import copy2

log = logging.getLogger(__package__+".yum") # maybe I should rename this..

# TODO: add --urlgrabdebug to enable this... or something
#yum.urlgrabber.grabber.set_logger(logging.getLogger(__package__+".urlgrab"))

def mirrorlist(repo, arch='$basearch'):
    return mirrormanager + '?repo=%s&arch=%s' % (repo, arch)

def raise_exception(failobj):
    raise failobj.exception

pluginpath = []
def yum_plugin_for_exc():
    import sys, traceback
    tb_files = [i[0] for i in traceback.extract_tb(sys.exc_info()[2])]
    log.debug("checking traceback files: %s", tb_files)
    log.debug("plugin path is %s", pluginpath)
    for f in tb_files:
        for p in pluginpath:
            if f.startswith(p):
                return f
    return None

def init_keyring(gpgdir):
    # set up gpgdir
    if not os.path.isdir(gpgdir):
        log.debug("creating gpgdir %s", gpgdir)
        os.makedirs(gpgdir, 0o700)
    else:
        os.chmod(gpgdir, 0o700)
    os.environ['GNUPGHOME'] = gpgdir

def import_key(keydata, hexkeyid, gpgdir):
    log.debug("importing key %s", hexkeyid.lower())
    yum.misc.import_key_to_pubring(keydata, hexkeyid,
                                   gpgdir=gpgdir, make_ro_copy=False)

def list_keyring(gpgdir):
    return [yum.misc.keyIdToRPMVer(int(k, 16))
            for k in yum.misc.return_keyids_from_pubring(gpgdir)]


class UpgradeDownloader(yum.YumBase):
    '''Yum-based downloader class. Based roughly on AnacondaYum.'''
    def __init__(self, version=None, cachedir=cachedir, cacheonly=False):
        # TODO: special handling for version='test' where we just synthesize
        #       a bunch of fake RPMs with interesting properties
        log.info("UpgradeDownloader(version=%s,cachedir=%s)",version,cachedir)
        yum.YumBase.__init__(self)
        self.use_txmbr_in_callback = True
        self.preconf.debuglevel = -1
        self.preconf.enabled_plugins = enabled_plugins
        self.preconf.disabled_plugins = disabled_plugins
        self.version = version
        if version:
            self.preconf.releasever = version
        self.cacheonly = cacheonly
        self.prerepoconf.cachedir = cachedir
        self.prerepoconf.cache = cacheonly
        log.debug("prerepoconf.cache=%i", self.prerepoconf.cache)
        self.instrepoid = None
        self.disabled_repos = []
        self._treeinfo = None
        self.prerepoconf.failure_callback = raise_exception
        self._repoprogressbar = None
        # TODO: locking to prevent multiple instances
        self.verbose_logger = log

    def _getConfig(self):
        firstrun = hasattr(self, 'preconf')
        conf = yum.YumBase._getConfig(self)
        if firstrun:
            # override some of yum's defaults
            conf.disable_excludes = ['all']
            conf.cache = self.cacheonly
            conf.deltarpm = 0
            log.debug("conf.cache=%i", conf.cache)
        return conf

    def doPluginSetup(self, *args, **kwargs):
        yum.YumBase.doPluginSetup(self, *args, **kwargs)
        # Now that plugins have been set up, let's save some info about them
        global pluginpath
        pluginpath = self.plugins.searchpath
        log.info("enabled plugins: %s", self.plugins._plugins.keys())

    def add_repo(self, repoid, baseurls=[], mirrorlist=None, noverifyssl=False, **kwargs):
        '''like add_enable_repo, but doesn't do initial repo setup and doesn't
        make unnecessary changes'''
        r = yum.yumRepo.YumRepository(repoid)
        r.name = repoid
        r.base_persistdir = cachedir
        r.basecachedir = cachedir
        r.cache = self.cacheonly
        r.callback = kwargs.get('callback') or self._repoprogressbar
        r.failure_obj = raise_exception
        r.mirror_failure_obj = raise_exception
        r.baseurl = [varReplace(u, self.conf.yumvar) for u in baseurls if u]
        r.sslverify = not(noverifyssl)
        if mirrorlist:
            r.mirrorlist = varReplace(mirrorlist, self.conf.yumvar)
        self._repos.add(r)
        self._repos.enableRepo(repoid)

    def setup_repos(self, callback=None, progressbar=None, repos=[], noverifyssl=False):
        '''Return a list of repos that had problems setting up.'''
        # These will set up progressbar and callback when we actually do setup
        self.prerepoconf.progressbar = progressbar
        self.prerepoconf.callback = callback
        self._repoprogressbar = progressbar

        # TODO invalidate cache if the version doesn't match previous version
        log.info("checking repos")

        # Add default instrepo (and its key) if needed
        if self.instrepoid is None:
            self.instrepoid = 'default-installrepo'
            # FIXME: hardcoded and RedHat-specific
            mirrorurl = mirrorlist('redhat-install-$releasever')
            repos.append(('add', '%s=@%s' % (self.instrepoid, mirrorurl)))
            repos.append(('gpgkey', '%s=%s' % (self.instrepoid, defaultkey)))

        # We need to read .repo files before we can enable/disable them, so:
        self.repos # implicit repo setup! ha ha! what fun!

        self.repos.disableRepo('*')

        # user overrides to enable/disable repos.
        # NOTE: will raise YumBaseError if there are problems
        for action, repo in repos:
            if action == 'enable':
                self.repos.enableRepo(repo)
            elif action == 'disable':
                self.repos.disableRepo(repo)
            elif action == 'add':
                (repoid, url) = repo.split('=',1)
                if url[0] == '@':
                    self.add_repo(repoid, mirrorlist=url[1:], noverifyssl=noverifyssl)
                else:
                    self.add_repo(repoid, baseurls=[url], noverifyssl=noverifyssl)
                if self.conf.proxy:
                    repo = self.repos.getRepo(repoid)
                    repo.proxy = self.conf.proxy
                    repo.proxy_username = self.conf.proxy_username
                    repo.proxy_password = self.conf.proxy_password

        # add GPG keys *after* the repos are created
        for action, repo in repos:
            if action == 'gpgkey':
                (repoid, keyurl) = repo.split('=',1)
                repo = self.repos.getRepo(repoid)
                repo.gpgkey.append(varReplace(keyurl, self.conf.yumvar))
                repo.gpgcheck = True

        # check enabled repos
        for repo in self.repos.listEnabled():
            try:
                md_types = repo.repoXML.fileTypes()
            except yum.Errors.RepoError:
                log.info("can't find valid repo metadata for %s", repo.id)
                repo.disable()
                self.disabled_repos.append(repo.id)
            else:
                log.info("repo %s seems OK" % repo.id)

            # Disable gpg key checking for the repos, if requested
            if self._override_sigchecks:
                repo._override_sigchecks = True

        log.debug("repos.cache=%i", self.repos.cache)

        return self.disabled_repos

    def save_repo_configs(self):
        '''save repo configuration files for later use'''
        repodir = os.path.join('/etc/yum.repos.d')
        mkdir_p(repodir)
        for repo in self.repos.listEnabled():
            repofile = os.path.join(repodir, "redhat-upgrade-%s.repo" % repo.id)
            try:
                with open(repofile, 'w') as f:
                    f.write("[redhat-upgrade-%s]\n" % repo.id)
                    f.write("name=Upgrade - %s\n" % repo.id)
                    f.write("enabled=1\n")
                    f.write("skip_if_unavailable=True\n")
                    f.write("sslverify=%s\n" % repo.sslverify)
                    if repo.gpgcheck:
                        f.write("gpgcheck=1\n")
                    else:
                        f.write("gpgcheck=0\n")
                    if repo.gpgkey:
                        f.write("gpgkey=%s\n" % " ".join(repo.gpgkey))
                    if repo.mirrorlist:
                        f.write("mirrorlist=%s\n" % repo.mirrorlist)
                    elif repo.metalink:
                        f.write("metalink=%s\n" % repo.metalink)
                    elif repo.baseurl:
                        f.write("baseurl=%s\n" % repo.baseurl[0])
                    else:
                        log.error("repo %s has no baseurl, mirrorlist or metalink", repo.id)
                        f.close()
                        os.unlink(repofile)
                        continue
            except IOError as e:
                log.warn("couldn't write repofile for %s: %s", repo.id, str(e))

    # NOTE: could raise RepoError if metadata is missing/busted
    def build_update_transaction(self, callback=None):
        log.info("looking for updates")
        self.dsCallback = callback
        self.update()
        (rv, msgs) = self.buildTransaction(unfinished_transactions_check=False)
        # NOTE: self.po_with_problems is now a list of (po1, po2, errmsg) tuples
        log.info("buildTransaction returned %i", rv)
        for m in msgs:
            log.info("    %s", m)
        # NOTE: we ignore errors, as anaconda did before us.
        self.dsCallback = None
        return [t.po for t in self.tsInfo.getMembers()
                     if t.ts_state in ("i", "u")]

    def find_packages_without_updates(self):
        '''packages on the local system that aren't being updated/obsoleted'''
        remove = self.tsInfo.getMembersWithState(output_states=TS_REMOVE_STATES)
        return set(p for p in self.rpmdb if p not in remove)

    def describe_transaction_problems(self):
        problems = []

        def find_replacement(po):
            for tx in self.tsInfo.getMembers(po.pkgtup):
                # XXX multiple replacers?
                for otherpo, rel in tx.relatedto:
                    if rel in ('obsoletedby', 'updatedby'):
                        return po, otherpo
                    if rel in ('obsoletes', 'updates'):
                        return otherpo, po
            if po in self.rpmdb:
                return po, None
            else:
                return None, po

        def format_replacement(po):
            oldpkg, newpkg = find_replacement(po)
            if oldpkg and newpkg:
                return "%s (replaced by %s)" % (oldpkg, newpkg)
            elif oldpkg:
                return "%s (no replacement)" % oldpkg
            elif newpkg:
                return "%s (new package)" % newpkg

        done = set()
        for pkg1, pkg2, err in self.po_with_problems:
            if (pkg1,pkg2) in done:
                continue
            if pkg2 is not None:
                problems.append("%s requires %s" % (format_replacement(pkg1),
                                                    format_replacement(pkg2)))
            else:
                problems.append("%s has depsolving problems" % \
                                                    format_replacement(pkg1))
            done.add((pkg1,pkg2))

        return problems

    def download_packages(self, pkgs, callback=None):
        # Verifying a full upgrade payload of ~2000 pkgs takes a good 90-120
        # seconds with no callback. Unacceptable!
        # So: here we have our own verifyPkg loop, with callback.
        # The results get cached, so when yum does it again in the real
        # _downloadPackages function it's a negligible delay.
        localpkgs = [p for p in pkgs if os.path.exists(p.localPkg())]
        total = len(localpkgs)
        # XXX: multithreading?
        for num, p in enumerate(localpkgs, 1):
            local = p.localPkg()
            if hasattr(callback, "verify") and callable(callback.verify):
                callback.verify(num, total, local, None)
            ok = self.verifyPkg(local, p, False) # result will be cached by yum
        log.info("beginning package download...")
        updates = self._downloadPackages(callback)

        # Handle _downloadPackages returning None instead of an empty list
        if updates is None:
            updates = []

        if set(updates) != set(pkgs):
            log.debug("differences between requested pkg set and downloaded:")
            for p in set(pkgs).difference(updates):
                log.debug("  -%s", p)
            for p in set(updates).difference(pkgs):
                log.debug("  +%s", p)
        # check signatures of downloaded packages
        if updates:
            self._checkSignatures(updates, callback)

        # store RHSM productid certificates
        # (this code is inspired by is taken from subscription_manager.productid)
        used_repos = set([p.repoid for p in updates])
        for repo in self.repos.listEnabled():
            # skip repos we have not used to download any package
            if repo.id not in used_repos:
                continue

            # skip repos we don't have productid for
            try:
                fn = repo.retrieveMD('productid')
                if fn.endswith('.gz'):
                    f = GzipFile(fn)
                else:
                    f = open(fn)
                try:
                    pem = f.read()
                    cert = create_from_pem(pem)
                finally:
                    f.close()
            except yum.Errors.RepoMDError, e:
                log.warn("Error loading productid metadata for %s." % repo)
                continue
            except Exception, e:
                log.warn("Error loading productid metadata for %s." % repo)
                log.exception(e)
                continue
            if cert is None:
                continue

            product = cert.products[0]

            log.debug("product cert: %s repo: %s" % (product.id, repo.id))

            # temporarily use different directory for storing certificates
            packagedir = '/root/preupgrade/productid'
            mkdir_p(packagedir)
            path = os.path.join(packagedir, '%s.pem' % product.id)
            cert.write(path)
            log.info("Downloaded product cert %s: %s %s" % (product.id, product.name, cert.path))

    def clean_cache(self, keepfiles):
        log.info("checking for unneeded rpms in cache")
        # Find all the packages in the caches (not on media though)
        localpkgs = set(f for r in self.repos.listEnabled() if not r.mediaid
                          for f in listdir(r.pkgdir) if f.endswith(".rpm"))
        for f in localpkgs.difference(keepfiles):
            try:
                log.debug("removing %s", f)
                os.remove(f)
            except IOError as e:
                log.info("failed to remove %s", f)
        # TODO remove dirs that don't belong to any repo

    @property
    def instrepo(self):
        return self.repos.getRepo(self.instrepoid)

    @property
    def treeinfo(self):
        if self._treeinfo is None:
            mkdir_p(cachedir)
            outfile = os.path.join(cachedir, '.treeinfo')
            if self.cacheonly:
                log.debug("using cached .treeinfo %s", outfile)
                self._treeinfo = Treeinfo(outfile)
            else:
                log.debug("fetching .treeinfo from repo '%s'", self.instrepoid)
                if os.path.exists(outfile):
                    os.remove(outfile)
                try:
                    fn = self.instrepo.grab.urlgrab('.treeinfo', outfile,
                                                    reget=None)
                except Exception:
                    try:
                        fn = self.instrepo.grab.urlgrab('treeinfo', outfile,
                                                        reget=None)
                    except Exception as e:
                        log.error("Error downloading .treeinfo or treeinfo"
                                  "from repo %s: %s" % (self.instrepoid, e))
                        fn = None
                self._treeinfo = Treeinfo(fn)
                log.debug(".treeinfo saved at %s", fn)
            self._treeinfo.checkvalues()
        return self._treeinfo

    def download_boot_images(self, arch=None):
        # helper function to grab and checksum image files listed in .treeinfo
        def grab_and_check(imgarch, imgtype, outpath):
            relpath = self.treeinfo.get_image(imgarch, imgtype)
            log.debug("grabbing %s %s", imgarch, imgtype)
            log.info("downloading %s to %s", relpath, outpath)
            if self.treeinfo.checkfile(outpath, relpath):
                log.debug("file already exists and checksum OK")
                return outpath
            def checkfile(cb):
                log.debug("checking %s", relpath)
                if not self.treeinfo.checkfile(cb.filename, relpath):
                    log.info("checksum doesn't match - retrying")
                    raise yum.URLGrabError(-1)
            return self.instrepo.grab.urlgrab(relpath, outpath,
                                              checkfunc=checkfile,
                                              reget=None,
                                              copy_local=True)

        # download the images
        try:
            if not arch:
                arch = self.treeinfo.get('general', 'arch')
            kernel = grab_and_check(arch, 'kernel', kernelpath)
            # cache the initrd somewhere so we don't have to fetch it again
            # if it gets modified later.
            cacheinitrd = os.path.join(cachedir, os.path.basename(initrdpath))
            initrd = grab_and_check(arch, 'upgrade', cacheinitrd)
            # copy the downloaded initrd to the target path
            try:
                copy2(initrd, initrdpath)
            except IOError as e:
                print _("Copying initrd to '%s' failed:\n%s") % (initrdpath, e)
                raise SystemExit(1)
            initrd = initrdpath
        except TreeinfoError as e:
            raise YumBaseError(_("invalid data in .treeinfo: %s") % str(e))
        except yum.URLGrabError as e:
            err = e.strerror
            if e.errno == 256:
                err += "\n" + _("Last error was: %s") % e.errors[-1][1]
            raise YumBaseError(_("couldn't get boot images: %s") % err)
        except KeyboardInterrupt:
            # if an IOError occurs while writing the file to disk, F17
            # urlgrabber actually raises *KeyboardInterrupt* for some reason.
            # But urlgrabber.__version__ hasn't been changed since F12, so:
            if not hasattr(yum.urlgrabber.grabber, 'exception2msg'): # <=F17
                raise KeyboardInterrupt(_("or possible error writing file"))
            else:
                # The exception actually was a KeyBoardInterrupt, re-raise it
                raise

        # Save kernel/initrd info so we can clean it up later
        with Config(upgradeconf) as conf:
            conf.set("boot", "kernel", kernel)
            conf.set("boot", "initrd", initrd)

        return kernel, initrd

    def _checkSignatures(self, pkgs, callback):
        '''check the package signatures and get keys if needed.
           works like YumBase._checkSignatures() except it only uses our
           special automatic _GPGKeyCheck to import untrusted keys.'''
        for po in pkgs:
            result, errmsg = self.sigCheckPkg(po)
            if result == 0:
                continue
            elif result == 1:
                keycheck = lambda info: self._GPGKeyCheck(info, callback)
                self.getKeyForPackage(po, fullaskcb=keycheck)
            else:
                raise yum.Errors.YumGPGCheckError(errmsg)

    def _GPGKeyCheck(self, info, callback=None):
        '''special key importer: import trusted keys automatically'''
        if info['keyurl'].startswith("file://"):
            keyfile = info['keyurl'][7:]
        else:
            return False
        po = info['po']
        log.info("repo '%s' wants to import key %s", po.repoid, keyfile)
        if self.check_keyfile(keyfile):
            log.info("key was installed by signed, trusted package - importing")
            return True
        else:
            log.info("no automatic trust for key %s")
            if hasattr(callback, "userconfirm"):
                return callback.userconfirm()
            return False

    def check_keyfile(self, keyfile):
        '''
        If a keyfile was installed by a package that was signed with a trusted
        key (and the key hasn't been modified or tampered with), we can assume
        that the key is trustworthy.

        This is kind of a roundabout way to establish trust between the two
        keys. It'd be a lot more straightforward if we just signed the new
        release key with the old release key - "If you trust this, you can
        trust this too.."
        '''
        if keyfile.startswith('file://'):
            keyfile = keyfile[7:]
        # did the key come from a package?
        keypkgs = self.rpmdb.searchFiles(keyfile)
        log.info("checking keyfile %s", keyfile)
        if keypkgs:
            keypkg = sorted(keypkgs)[-1]
            log.debug("keyfile owned by package %s", keypkg.nevr)
        if not keypkgs:
            log.info("REJECTED: %s does not belong to any package")
            return False

        # was that package signed?
        hdr = keypkg.returnLocalHeader()
        if hdr.sigpgp or hdr.siggpg:
            sigdata = hdr.sigpgp or hdr.siggpg
            siginfo = yum.pgpmsg.decode(sigdata)[0]
            (keyid,) = struct.unpack('>Q', siginfo.key_id())
            hexkeyid = yum.misc.keyIdToRPMVer(keyid)
            log.debug("package was signed with key %s", hexkeyid)
        else:
            log.info("REJECTED: %s was unsigned", keypkg.nevr)
            return False

        # do we trust the key that signed it?
        if yum.misc.keyInstalled(self.ts, keyid, 0) >= 0:
            log.debug("key %s is trusted by rpm", hexkeyid)
        else:
            log.info("REJECTED: key %s is not trusted by rpm", hexkeyid)
            return False

        # has the key been tampered with?
        problems = keypkg.verify([keyfile]).get(keyfile, [])
        if problems:
            log.info("REJECTED: keyfile does not match packaged file (%s)",
                     keyfile, " ".join(p.type for p in problems))
            return False

        # everything checks out OK!
        return True

    def _setup_keyring(self, gpgdir):
        # set up a fresh gpgdir
        rm_rf(gpgdir)
        init_keyring(gpgdir)

        # import trusted keys from rpmdb
        log.debug("checking rpmdb trusted keys")
        pubring_keys = list_keyring(gpgdir)
        for hdr in self.ts.dbMatch('name', 'gpg-pubkey'):
            if hdr.version not in pubring_keys:
                import_key(hdr.description, hdr.version, gpgdir)
            else:
                log.debug("key %s is already in keyring", hdr.version)

        # check instrepo keys to see if they're trustworthy
        log.info("checking GPG keys for instrepo")
        for k in self.instrepo.gpgkey:
            if self.check_keyfile(k):
                keys = self._retrievePublicKey(k) # XXX getSig?
                for info in keys:
                    import_key(info['raw_key'], info['hexkeyid'], gpgdir)

    def check_signed_file(self, signedfile, outfile, gpgdir=cachedir+'/gpgdir'):
        '''
        uses the keys trusted by RPM to verify signedfile.
        writes the resulting plaintext to outfile.
        returns a list of unicode strings describing any errors in verification.
        if the list is empty, the verification was successful.

        It'd be great if RPM could do this for us, since it already has all the
        keys imported and has its own signature verification code, but AFAICT
        it doesn't (at least not in any way reachable from Python), so..
        '''
        if self._override_sigchecks:
            return []

        # set up our own GPG keyring containing all the trusted keys
        self._setup_keyring(gpgdir)

        # verify the signed file, writing plaintext to outfile
        with open(signedfile) as inf:
            with open(outfile, 'w') as outf:
                ctx = gpgme.Context()
                sigresults = ctx.verify(inf, None, outf)
        # return a list of error messages. if it's empty, we're OK.
        # NOTE: this is enough detail for current use cases, but it's very
        # possible we'll want/need to just return sigresults and let the caller
        # sort out the details..
        return [sig.status.message for sig in sigresults if not (
                     sig.summary & gpgme.SIGSUM_VALID and
                     sig.validity >= gpgme.VALIDITY_FULL)]



# upgrade.py - test the upgrade transaction using RPM
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>


# For the sake of simplicity, we don't bother with yum here.
import rpm
from rpm._rpm import ts as TransactionSetCore

import os, tempfile
from threading import Thread

import logging
log = logging.getLogger(__package__+'.upgrade')

from . import _
from .util import df, hrsize

class TransactionSet(TransactionSetCore):
    flags = TransactionSetCore._flags
    vsflags = TransactionSetCore._vsflags
    color = TransactionSetCore._color

    def run(self, callback, data, probfilter):
        log.debug('ts.run()')
        rv = TransactionSetCore.run(self, callback, data, probfilter)
        problems = self.problems()
        if rv != rpm.RPMRC_OK and problems:
            raise TransactionError(problems)
        return rv

    def check(self, *args, **kwargs):
        TransactionSetCore.check(self, *args, **kwargs)
        # NOTE: rpm.TransactionSet throws out all problems but these
        return [p for p in self.problems()
                  if p.type in (rpm.RPMPROB_CONFLICT, rpm.RPMPROB_REQUIRES)]

    def add_install(self, path, key=None, upgrade=False):
        log.debug('add_install(%s, %s, upgrade=%s)', path, key, upgrade)
        if key is None:
            key = path
        with open(path) as fileobj:
            retval, header = self.hdrFromFdno(fileobj)
        if retval != rpm.RPMRC_OK:
            raise rpm.error("error reading package header")
        if not self.addInstall(header, key, upgrade):
            raise rpm.error("adding package to transaction failed")

    def __del__(self):
        self.closeDB()

probtypes = { rpm.RPMPROB_NEW_FILE_CONFLICT : _('file conflicts'),
              rpm.RPMPROB_FILE_CONFLICT : _('file conflicts'),
              rpm.RPMPROB_OLDPACKAGE: _('older package(s)'),
              rpm.RPMPROB_DISKSPACE: _('insufficient disk space'),
              rpm.RPMPROB_DISKNODES: _('insufficient disk inodes'),
              rpm.RPMPROB_CONFLICT: _('package conflicts'),
              rpm.RPMPROB_PKG_INSTALLED: _('package already installed'),
              rpm.RPMPROB_REQUIRES: _('broken dependencies'),
              rpm.RPMPROB_BADARCH: _('package for incorrect arch'),
              rpm.RPMPROB_BADOS: _('package for incorrect os'),
            }

# --- stuff for doing useful summaries of big sets of problems

probattrs = ('type', 'pkgNEVR', 'altNEVR', 'key', '_str', '_num')
def prob2dict(p):
    dict = {}
    for f in probattrs:
        dict[f] = getattr(p, f)
    return dict

class ProblemSummary(object):
    def __init__(self, probtype, problems):
        self.type = probtype
        self.problems = [p for p in problems if p.type == self.type]
        self.desc = probtypes.get(probtype)
        self.details = self.get_details()

    def get_details(self):
        return None

    def format_details(self):
        raise NotImplementedError

    def _log_probs(self):
        for p in self.problems:
            log.debug('%s -> "%s"', prob2dict(p), p)

    def __str__(self):
        if self.details:
            return "\n  ".join([self.desc+':'] + self.format_details())
        else:
            return self.desc

class DiskspaceProblemSummary(ProblemSummary):
    def get_details(self):
        needs = dict()
        for p in self.problems:
            (mnt, size) = (p._str, p._num)
            if size > needs.get(mnt,0):
                needs[mnt] = size
        return needs

    def format_details(self):
        return [_("%s needs %s more free space") % (mnt, hrsize(size))
                 for (mnt,size) in self.details.iteritems()]

class DepProblemSummary(ProblemSummary):
    def get_details(self):
        self._log_probs()
        pkgprobs = dict()
        # pkgprobs['installedpkg'] = [req1, req2, ...]
        for p in self.problems:
            thispkg, req = p.pkgNEVR, p.altNEVR.split()[1]
            if thispkg not in pkgprobs:
                pkgprobs[thispkg] = set()
            pkgprobs[thispkg].add(req)
        return pkgprobs

    def format_details(self):
        return [_("%s requires %s") % (pkg, ", ".join(pkgprob))
                 for (pkg, pkgprob) in self.details.iteritems()]

# If there is no handler for a type of problem, just return the
# rpmProblemString result for the problems
class GenericProblemSummary(ProblemSummary):
    def format_details(self):
        return [str(p) for p in self.problems]

probsummary = { rpm.RPMPROB_DISKSPACE: DiskspaceProblemSummary,
                rpm.RPMPROB_REQUIRES:  DepProblemSummary,
              }


def summarize_problems(problems):
    summaries = []
    for t in set(p.type for p in problems):
        summarize = probsummary.get(t, GenericProblemSummary) # get the summarizer
        summaries.append(summarize(t, problems))       # summarize the problem
    return summaries

class TransactionError(Exception):
    def __init__(self, problems):
        self.problems = problems
        self.summaries = summarize_problems(problems)

def pipelogger(pipe, level=logging.INFO):
    logger = logging.getLogger(__package__+".rpm")
    logger.info("opening pipe")
    with open(pipe, 'r') as fd:
        for line in fd:
            if line.startswith('D: '):
                logger.debug(line[3:].rstrip())
            else:
                logger.log(level, line.rstrip())
        logger.info("got EOF")
    logger.info("exiting")

logging_to_rpm = {
    logging.DEBUG:      rpm.RPMLOG_DEBUG,
    logging.INFO:       rpm.RPMLOG_INFO,
    logging.WARNING:    rpm.RPMLOG_WARNING,
    logging.ERROR:      rpm.RPMLOG_ERR,
    logging.CRITICAL:   rpm.RPMLOG_CRIT,
}

class RPMUpgrade(object):
    def __init__(self, root='/', logpipe=True, rpmloglevel=logging.INFO):
        self.root = root
        self.ts = None
        self.logpipe = None
        rpm.setVerbosity(logging_to_rpm[rpmloglevel])
        if logpipe:
            self.logpipe = self.openpipe()

    def setup_transaction(self, pkgfiles, check_fatal=False):
        log.debug("starting")
        # initialize a transaction set
        self.ts = TransactionSet(self.root, rpm._RPMVSF_NOSIGNATURES)
        if self.logpipe:
            self.ts.scriptFd = self.logpipe.fileno()
        # populate the transaction set
        for pkg in pkgfiles:
            try:
                self.ts.add_install(pkg, upgrade=True)
            except rpm.error as e:
                log.warn('error adding pkg: %s', e)
                # TODO: error callback
        log.debug('ts.check()')
        problems = self.ts.check() or []
        if problems:
            log.info("problems with transaction check:")
            for p in problems:
                log.info(p)
            if check_fatal:
                raise TransactionError(problems=problems)

        log.debug('ts.order()')
        self.ts.order()
        log.debug('ts.clean()')
        self.ts.clean()
        log.debug('transaction is ready')
        if problems:
            return TransactionError(problems=problems)

    def openpipe(self):
        log.debug("creating log pipe")
        pipefile = tempfile.mktemp(prefix='rpm-log-pipe.')
        os.mkfifo(pipefile, 0600)
        log.debug("starting logging thread")
        pipethread = Thread(target=pipelogger, name='pipelogger',
                                 args=(pipefile,))
        pipethread.daemon = True
        pipethread.start()
        log.debug("opening log pipe")
        pipe = open(pipefile, 'w')
        rpm.setLogFile(pipe)
        return pipe

    def closepipe(self):
        log.debug("closing log pipe")
        rpm.setVerbosity(rpm.RPMLOG_WARNING)
        rpm.setLogFile(None)
        if self.ts:
            self.ts.scriptFd = None
        self.logpipe.close()
        os.remove(self.logpipe.name)
        self.logpipe = None

    def run_transaction(self, callback):
        assert callable(callback.callback)
        probfilter = ~rpm.RPMPROB_FILTER_DISKSPACE
        rv = self.ts.run(callback.callback, None, probfilter)
        if rv != 0:
            log.info("ts completed with problems - code %u", rv)
        return rv

    def test_transaction(self, callback):
        self.ts.flags = rpm.RPMTRANS_FLAG_TEST
        try:
            return self.run_transaction(callback)
        finally:
            self.ts.flags &= ~rpm.RPMTRANS_FLAG_TEST

    def __del__(self):
        if self.logpipe:
            self.closepipe()



#!/usr/bin/python

from distutils.core import setup, Command
from distutils.util import convert_path
from distutils.command.build_scripts import build_scripts
from distutils import log

import os
from os.path import join, basename
from subprocess import check_call

class Gettext(Command):
    description = "Use po/POTFILES.in to generate po/<name>.pot"
    user_options = []
    def initialize_options(self):
        self.encoding = 'UTF-8'
        self.po_dir = 'po'
        self.add_comments = True

    def finalize_options(self):
        pass

    def _xgettext(self, opts):
        name = self.distribution.get_name()
        version = self.distribution.get_version()
        email = self.distribution.get_author_email()
        cmd = ['xgettext', '--default-domain', name, '--package-name', name,
               '--package-version', version, '--msgid-bugs-address', email,
               '--from-code', self.encoding,
               '--output', join(self.po_dir, name + '.pot')]
        if self.add_comments:
            cmd.append('--add-comments')
        check_call(cmd + opts)

    def run(self):
        self._xgettext(['-f', 'po/POTFILES.in'])

class Msgfmt(Command):
    description = "Generate po/*.mo from po/*.po"
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        po_dir = 'po'
        for po in os.listdir(po_dir):
            po = join(po_dir, po)
            if po.endswith('.po'):
                mo = po[:-3]+'.mo'
                check_call(['msgfmt', '-vv', po, '-o', mo])

class BuildScripts(build_scripts):
    def run(self):
        build_scripts.run(self)
        for script in self.scripts:
            script = convert_path(script)
            outfile = join(self.build_dir, basename(script))
            if os.path.exists(outfile) and outfile.endswith(".py"):
                newfile = outfile[:-3] # drop .py
                log.info("renaming %s -> %s", outfile, basename(newfile))
                os.rename(outfile, newfile)

setup(name="redhat-upgrade-tool",
      version="0.7.48",
      description="Red Hat Upgrade",
      long_description="",
      author="Will Woods",
      author_email="wwoods@redhat.com",
      url="https://github.com/dashea/redhat-upgrade-tool",
      download_url="https://github.com/dashea/redhat-upgrade-tool/downloads",
      license="GPLv2+",
      packages=["redhat_upgrade_tool"],
      scripts=["redhat-upgrade-tool.py"],
      cmdclass={
        'gettext': Gettext,
        'msgfmt': Msgfmt,
        'build_scripts': BuildScripts,
        }
      )



# sysprep.py - utility functions for system prep
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import glob
import os
import re
from shutil import copy2

from . import _
from . import cachedir, packagedir, packagelist, update_img_dir
from . import upgradeconf, upgradelink, upgraderoot
from . import boot
from .media import write_prep_mount
from .util import listdir, mkdir_p, rm_f, rm_rf, is_selinux_enabled, kernelver
from .conf import Config
from .repofile import RepoFileParser

import logging
log = logging.getLogger(__package__ + ".sysprep")

upgrade_prep_dir = packagedir + "/upgrade-prep"
repodir = '/etc/yum.repos.d'


def setup_cleanup_post():
    '''Set a flag in upgrade.conf to be read by preupgrade-assistant,
       signalling it to cleanup old packages in the post scripts.'''

    with Config(upgradeconf) as conf:
        conf.set('postupgrade', 'cleanup', 'True')


def link_pkgs(pkgs):
    '''link the named pkgs into packagedir, overwriting existing files.
       also removes any .rpm files in packagedir that aren't in pkgs.
       finally, write a list of packages to upgrade and a list of dirs
       to clean up after successful upgrade.'''

    log.info("linking required packages into packagedir")
    log.info("packagedir = %s", packagedir)
    mkdir_p(packagedir)

    pkgbasenames = set()
    for pkg in pkgs:
        pkgpath = pkg.localPkg()
        if pkg.remote_url.startswith("file://"):
            pkgbasename = "media/%s" % pkg.relativepath
            pkgbasenames.add(pkgbasename)
            continue
        if not os.path.exists(pkgpath):
            log.warning("%s missing", pkgpath)
            continue
        pkgbasename = os.path.basename(pkgpath)
        pkgbasenames.add(pkgbasename)
        target = os.path.join(packagedir, pkgbasename)
        if os.path.exists(target) and os.lstat(pkgpath) == os.lstat(target):
            log.info("%s already in packagedir", pkgbasename)
            continue
        else:
            if os.path.isdir(target):
                log.info("deleting weirdo directory named %s", pkgbasename)
                rm_rf(target)
            elif os.path.exists(target):
                os.remove(target)
            try:
                os.link(pkgpath, target)
            except OSError as e:
                if e.errno == 18:
                    copy2(pkgpath, target)
                else:
                    raise

    # remove spurious / leftover RPMs
    for f in os.listdir(packagedir):
        if f.endswith(".rpm") and f not in pkgbasenames:
            os.remove(os.path.join(packagedir, f))

    # write packagelist
    with open(packagelist, 'w') as outf:
        outf.writelines(p + '\n' for p in pkgbasenames)

    # write cleanup data
    with Config(upgradeconf) as conf:
        # packagedir should probably be last, since it contains upgradeconf
        cleanupdirs = [cachedir, packagedir]
        conf.set("cleanup", "dirs", ';'.join(cleanupdirs))


def setup_upgradelink():
    log.info("setting up upgrade symlink: %s->%s", upgradelink, packagedir)
    try:
        os.remove(upgradelink)
    except OSError:
        pass
    os.symlink(packagedir, upgradelink)


def setup_media_mount(mnt, iso):
    # make a "media" subdir where all the packages are
    mountpath = os.path.join(upgradelink, "media")
    log.info("setting up mount for %s at %s", mnt.dev, mountpath)
    mkdir_p(mountpath)
    # make a directory to place a unit
    mkdir_p(upgrade_prep_dir)
    # make a modified mnt entry that puts it at mountpath
    mediamnt = mnt._replace(rawmnt=mountpath)
    # finally, write out a systemd unit to mount media there
    unit = write_prep_mount(mediamnt, upgrade_prep_dir, iso)
    log.info("wrote %s", unit)


def setup_upgraderoot():
    if os.path.isdir(upgraderoot):
        log.info("upgrade root dir %s already exists", upgraderoot)
        return
    else:
        log.info("creating upgraderoot dir: %s", upgraderoot)
        os.makedirs(upgraderoot, 0755)


def prep_upgrade(pkgs):
    # put packages in packagedir (also writes packagelist)
    link_pkgs(pkgs)
    # make magic symlink
    setup_upgradelink()
    # make dir for upgraderoot
    setup_upgraderoot()


def init_is_systemd():
    try:
        return "systemd" in os.readlink("/sbin/init")
    except OSError:
        return False


def modify_bootloader(kernel, initrd):
    log.info("adding new boot entry")

    args = ["upgrade"]
    remove_args = ["rhgb", "quiet"]
    if init_is_systemd():
        args.append("systemd.unit=system-upgrade.target")
    else:
        args.append("init=/usr/libexec/upgrade-init")  # XXX hardcoded path :/

    if not is_selinux_enabled():
        args.append("selinux=0")
    else:
        # BLERG. SELinux enforcing will cause problems if the new policy
        # disallows something that the previous system did differently.
        # See https://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=896010
        args.append("enforcing=0")

    # Workaround for systemd not reaching sysinit.target in RHEL7 dracut
    args.append('rd.plymouth=0')
    args.append('plymouth.enable=0')

    # Leave the network names alone so that the RHEL-6 initscripts can
    # bring interfaces up
    args.append('net.ifnames=0')

    # Screen blanking just makes the screen broken
    args.append("consoleblank=0")

    boot.add_entry(kernel, initrd, banner=_("System Upgrade"), kargs=args,
                   remove_kargs=remove_args)


def prep_boot(kernel, initrd):
    # check for systems that need mdadm.conf
    if boot.need_mdadmconf():
        log.info("appending /etc/mdadm.conf to initrd")
        boot.initramfs_append_files(initrd, "/etc/mdadm.conf")

    # look for updates, and add them to initrd if found
    updates = []
    try:
        updates = list(listdir(update_img_dir))
    except (IOError, OSError) as e:
        log.info("can't list update img dir %s: %s", update_img_dir,
                 e.strerror)
    if updates:
        log.info("found updates in %s, appending to initrd", update_img_dir)
        boot.initramfs_append_images(initrd, updates)

    # make a dir in /lib/modules to hold a copy of the new kernel's modules
    # (the initramfs will copy/bind them into place when we reboot)
    kv = kernelver(kernel)
    if kv:
        moddir = os.path.join("/lib/modules", kv)
        log.info("creating module dir %s", moddir)
        mkdir_p(moddir)
    else:
        log.warn("can't determine version of kernel image '%s'", kernel)

    # set up the boot args
    modify_bootloader(kernel, initrd)


def reset_boot():
    '''reset bootloader to previous default and remove our boot entry'''
    conf = Config(upgradeconf)
    kernel = conf.get("boot", "kernel")
    if kernel:
        boot.remove_entry(kernel)


def remove_boot():
    '''remove boot images'''
    conf = Config(upgradeconf)
    kernel = conf.get("boot", "kernel")
    initrd = conf.get("boot", "initrd")
    if kernel:
        rm_f(kernel)
    if initrd:
        rm_f(initrd)


def remove_cache():
    '''remove our cache dirs'''
    log.info("Removing cache if present from previous run.")
    conf = Config(upgradeconf)
    cleanup = conf.get("cleanup", "dirs") or ''
    cleanup = set(cleanup.split(';'))
    cleanup.update([cachedir, packagedir])  # just to be sure
    for d in cleanup:
        log.info("removing %s", d)
        rm_rf(d)


def disable_old_repos():
    for repo in glob.glob(repodir + '/*.repo'):
        parser = RepoFileParser(repo)
        parser.set_option("enabled", "0\n# disabled by redhat-upgrade-tool",
                          "1")
        parser.write()


def misc_cleanup():
    log.info("removing symlink %s", upgradelink)
    rm_f(upgradelink)
    for d in (upgraderoot, upgrade_prep_dir):
        log.info("removing %s", d)
        rm_rf(d)

    log.info("removing repo files")
    for repo in glob.glob(repodir + '/redhat-upgrade-*.repo'):
        rm_rf(repo)

    # Un-disable the repo files
    for repo in glob.glob(repodir + '/*.repo'):
        with open(repo, 'r') as repofile:
            repodata = repofile.read()
        enabled_data = re.sub(r"enabled\s*=\s*0\n# disabled by"
                              " redhat-upgrade-tool", "enabled=1", repodata)
        if enabled_data != repodata:
            with open(repo, 'w') as repofile:
                repofile.write(enabled_data)


def modify_repos(args):
    """Rewrite the upgrade or iso repo configs for the post-reboot mount
    paths.
    """
    if args.device or args.iso:
        mountpath = os.path.join(upgradelink, "media")
        repopath = os.path.join(repodir, 'redhat-upgrade-{0}.repo'
                                .format(args.instrepo))
        parser = RepoFileParser(repopath)
        parser.set_option("baseurl", "file://{0}".format(mountpath))
        parser.write()



# textoutput - text output routines
# vim: set fileencoding=UTF-8:
#
# Copyright (C) 2012 Red Hat Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import os, sys, time

import rpm
sys.path.insert(0, '/usr/share/yum-cli')
from output import YumTextMeter, CacheProgressCallback, YumOutput

from .callback import *
from . import terminal as term

from . import _

import logging
log = logging.getLogger(__package__+".cli")


class SimpleProgress(object):
    def __init__(self, maxval, prefix="", barstyle='[=]', update_interval=0.3,
                 tty=sys.stdout):
        self.maxval = maxval
        self.curval = 0
        self.formatstr = "{0.prefix} {0.percent:2}% {0.bar}"
        self.barstyle = barstyle
        self.prefix = prefix
        # update screen at a certain interval
        self.tty = tty
        self.update_interval = update_interval
        self.screenupdate = 0

    @property
    def width(self):
        return term.size.cols or 80 # fallback for stupid terminals

    @property
    def percent(self):
        return int(100*self.curval / float(self.maxval))

    bar_fmt = "{l_br}{barchar:<{width}}{r_br}"
    @property
    def bar(self):
        otherstuff = self.formatstr.replace("{0.bar}","")
        barwidth = self.width - len(otherstuff.format(self)) - 2 # 2 brackets
        fillpart = barwidth * self.curval / self.maxval
        return self.bar_fmt.format(l_br=self.barstyle[0],
                                   barchar=self.barstyle[1] * fillpart,
                                   r_br=self.barstyle[2],
                                   width=barwidth)

    def __str__(self):
        return self.formatstr.format(self)

    def update(self, newval, forceupdate=False):
        now = time.time()
        self.curval = min(newval, self.maxval)
        if forceupdate or (now - self.screenupdate > self.update_interval):
            self.screenupdate = now
            self.tty.write("\r%s" % self)
            self.tty.flush()

    def finish(self):
        self.update(self.maxval, forceupdate=True)
        self.tty.write("\n")

class RepoProgress(YumTextMeter):
    pass

class RepoCallback(object):
    def __init__(self, prefix="repodata", tty=sys.stderr):
        self._pb = SimpleProgress(10, prefix=prefix, tty=tty)
    def progressbar(self, current, total, name=None):
        if name:
            self._pb.prefix = "repodata (%s)" % name
        self._pb.maxval = total
        self._pb.update(current)

class DepsolveCallback(DepsolveCallbackBase):
    def __init__(self, yumobj=None, tty=sys.stderr):
        DepsolveCallbackBase.__init__(self, yumobj)
        self.progressbar = None
        if yumobj and tty:
            self.progressbar = SimpleProgress(self.installed_packages, tty=tty,
                                              prefix=_("finding updates"))

    def pkgAdded(self, tup, mode):
        DepsolveCallbackBase.pkgAdded(self, tup, mode)
        if self.progressbar and mode == "ud":
            self.progressbar.update(self.mode_counter['ud'])

    def end(self):
        DepsolveCallbackBase.end(self)
        if self.progressbar:
            self.progressbar.finish()
            self.progressbar = None

class DownloadCallback(DownloadCallbackBase):
    def __init__(self, tty=sys.stderr):
        DownloadCallbackBase.__init__(self)
        self.bar = SimpleProgress(10, tty=tty, prefix=_("verify local files"))

    def verify(self, amount, total, filename, data):
        DownloadCallbackBase.verify(self, amount, total, filename, data)
        if self.bar.maxval != total:
            self.bar.maxval = total
        self.bar.update(amount)
        if amount == total:
            self.bar.finish()

    def userconfirm(self):
        def _raw_input_accepting_unicode(prompt):
            # When the terminal is set to a non-ASCII English language, the
            # translated prompt may contain Unicode chars - that results in
            # a python traceback within the raw_input python built-in function.
            # This wrapper fixes that.
            from yum.i18n import to_utf8
            return raw_input_saved(to_utf8(prompt))

        import __builtin__
        raw_input_saved = __builtin__.raw_input
        __builtin__.raw_input = _raw_input_accepting_unicode

        response = YumOutput().userconfirm()

        __builtin__.raw_input = raw_input_saved

        return response


class TransactionCallback(RPMTsCallback):
    def __init__(self, numpkgs=0, tty=sys.stderr, prefix="rpm"):
        RPMTsCallback.__init__(self)
        self.numpkgs = numpkgs
        self.donepkgs = 0
        self.progressbar = SimpleProgress(10, prefix="rpm transaction", tty=tty)
    def trans_start(self, amount, total, key, data):
        if amount != 6:
            log.warn("weird: trans_start() with amount != 6")
        self.progressbar.maxval = total
    def trans_progress(self, amount, total, key, data):
        self.progressbar.update(amount)
    def trans_stop(self, amount, total, key, data):
        self.progressbar.finish()

    def inst_open_file(self, amount, total, key, data):
        log.info("installing %s (%u/%u)", os.path.basename(key),
                                          self.donepkgs+1, self.numpkgs)
        if self.donepkgs == 0:
            self.progressbar.prefix = "rpm install"
            self.progressbar.maxval = self.numpkgs
        self.progressbar.update(self.donepkgs)
        return RPMTsCallback.inst_open_file(self, amount, total, key, data)

    def inst_close_file(self, amount, total, key, data):
        RPMTsCallback.inst_close_file(self, amount, total, key, data)
        self.donepkgs += 1

    def uninst_start(self, amount, total, key, data):
        log.info("cleaning %s", key)

    def __del__(self):
        if self.progressbar:
            self.progressbar.finish()
